'''Test functionality of PCR class of reaction module.'''
import os
import coral as cr
from coral import analysis, seqio, DNA, Primer
from nose.tools import assert_true, assert_raises


def test_basic():
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    # Test forward priming.
    seq = DNA('cgccagggttttcccagtcacgac')
    primer = Primer(seq, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    fwd_indices = [match[0] for match in fwd_matches]
    # fwd_lens = [match[1] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]
    # rev_lens = [match[1] for match in rev_matches]

    loc = template.locate(seq)

    assert_true(len(fwd_matches) == len(loc[0]))
    assert_true(len(rev_matches) == len(loc[1]))

    # Top strand matches
    for match in loc[0]:
        assert_true(match + len(seq) in fwd_indices)
    # Top strand matches
    for match in loc[1]:
        assert_true(match + len(seq) in rev_indices)

    # Test reverse priming
    seq = DNA('ACAAGAGAGATTGGGAAGGAAAGGATCA')
    primer = Primer(seq, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    fwd_indices = [match[0] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]

    loc = template.locate(seq)

    assert_true(len(fwd_indices) == len(loc[0]))
    assert_true(len(rev_indices) == len(loc[1]))
    for match in loc[0]:
        assert_true(match + len(seq) in fwd_indices)
    for match in loc[1]:
        assert_true(match + len(seq) in rev_indices)


def test_near_index():
    '''Test binding near index for circular templates.'''
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    template = template.circularize()
    seq = DNA('aggccctttcgtctcgcgcgttt')
    primer = Primer(seq, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    fwd_indices = [match[0] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]

    loc = template.locate(seq)

    print fwd_matches
    print loc[0]
    print loc[1]
    assert_true(len(fwd_matches) == len(loc[0]))
    assert_true(len(rev_matches) == len(loc[1]))
    for match in loc[0]:
        expected = match + len(seq)
        if expected > len(template):
            expected -= len(template)
        assert_true(expected in fwd_indices)
    for match in loc[1]:
        expected = match + len(seq)
        if expected > len(template):
            expected -= len(template)
        assert_true(expected in rev_indices)


def test_overhang():
    ''' test forward priming '''
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    seq = DNA('cgccagggttttcccagtcacgac')
    overhang = DNA('ggggggg')
    seq2 = overhang + seq
    primer = Primer(seq2, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches

    fwd_indices = [match[0] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]

    loc = template.locate(seq)

    assert_true(len(fwd_indices) == len(loc[0]))
    assert_true(len(rev_indices) == len(loc[1]))
    # FIXME: Add match length check for all these cases.
    for match in loc[0]:
        assert_true(match + len(seq) in fwd_indices)
    for match in loc[1]:
        assert_true(match + len(seq) in rev_indices)

    # Test forward priming.
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    seq = DNA('ACAAGAGAGATTGGGAAGGAAAGGATCA')
    overhang = DNA('ggggggg')
    seq2 = overhang + seq
    primer = Primer(seq2, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    fwd_indices = [match[0] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]

    loc = template.locate(seq)

    assert_true(len(fwd_indices) == len(loc[0]))
    assert_true(len(rev_indices) == len(loc[1]))

    for match in loc[0]:
        assert_true(match + len(seq) in fwd_indices)
    for match in loc[1]:
        assert_true(match + len(seq) in rev_indices)


def test_multiple_priming():
    ''' test multiple binding sites '''

    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    template = template.circularize()
    seq = DNA('cgccagggttttcccagtcacgac')
    template = template.linearize()
    template = template + seq + DNA("AGGCGTATGC") + seq
    template = (template + DNA("GGGGGGG") + seq.reverse_complement() +
                DNA("GGAAAG"))
    template = template.circularize()
    primer = Primer(seq, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    fwd_indices = [match[0] for match in fwd_matches]
    rev_indices = [match[0] for match in rev_matches]

    loc = template.locate(seq)

    assert_true(len(fwd_matches) == len(loc[0]))
    assert_true(len(rev_matches) == len(loc[1]))
    for match in loc[0]:
        assert_true(match + len(seq) in fwd_indices)
    for match in loc[1]:
        assert_true(match + len(seq) in rev_indices)


def test_no_priming():
    ''' test no priming '''

    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))
    seq = DNA('ggaggagggcggcgaggcgagcgacggaggggga')
    primer = Primer(seq, 50.6)
    matches = analysis.anneal(template, primer)
    fwd_matches, rev_matches = matches
    loc = template.locate(seq)
    assert_true(len(fwd_matches) == len(loc[0]))
    assert_true(len(rev_matches) == len(loc[1]))
    for match in loc[0]:
        assert_true(match + len(seq) in fwd_matches)
    for match in loc[1]:
        assert_true(match + len(seq) in rev_matches)


def test_min_primer_length():
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))

    # Test forward priming
    seq = DNA('cgccagggttttcccagtcacgac')
    seq = seq[:15]
    primer = Primer(seq, 50.6)
    assert_raises(analysis._sequence.anneal.PrimerLengthError, analysis.anneal,
                  template, primer, min_len=16)


def test_min_tm():
    current_path = os.path.dirname(__file__)
    template = seqio.read_dna(os.path.join(current_path,
                                           "pMODKan-HO-pACT1GEV.ape"))

    # Test forward priming
    # Tm should be ~40 C
    seq = DNA('CTTCTATCGAACAA')
    primer = Primer(seq, seq.tm())
    matches = analysis.anneal(template, primer, min_tm=60.0)
    assert_true(len(matches[0]) == 0)
    matches = analysis.anneal(template, primer, min_tm=30.0)
    assert_true(len(matches[0]) > 0)


def test_primer_larger_than_template():
    template = cr.design.random_dna(50)
    overhangs = [cr.design.random_dna(200), cr.DNA('')]
    expected = overhangs[0] + template
    primer1, primer2 = cr.design.primers(template, overhangs=overhangs,
                                         min_len=14)
    amplicon = cr.reaction.pcr(template, primer1, primer2)

    assert_true(expected == amplicon)



'''
Tests for the RNA sequence class.

'''

from coral import RNA
from nose.tools import assert_equal, assert_false, assert_true, assert_raises


class TestRNA(object):
    '''
    Testing class for sequence.RNA

    '''

    def __init__(self):
        self.test_rna = RNA('augc')

    def test_reverse_complement(self):
        assert_equal(str(self.test_rna.reverse_complement()), 'GCAU')

    def test_locate(self):
        assert_equal(self.test_rna.locate('au'), [0])
        assert_equal(self.test_rna.locate('gc'), [2])
        assert_equal(len(self.test_rna.locate('augg')), 0)

    def test_copy(self):
        assert_equal(self.test_rna, self.test_rna.copy())

    def test_getitem(self):
        assert_equal(str(self.test_rna[0]), 'A')
        assert_equal(str(self.test_rna[1]), 'U')
        assert_equal(str(self.test_rna[2]), 'G')
        assert_equal(str(self.test_rna[3]), 'C')
        assert_equal(str(self.test_rna[-1]), 'C')

    def test_delitem(self):
        copy0 = self.test_rna.copy()
        del copy0[0]
        assert_equal(str(copy0), 'UGC')
        copy1 = self.test_rna.copy()
        del copy1[1]
        assert_equal(str(copy1), 'AGC')
        copy2 = self.test_rna.copy()
        del copy2[2]
        assert_equal(str(copy2), 'AUC')
        copy3 = self.test_rna.copy()
        del copy3[3]
        assert_equal(str(copy3), 'AUG')
        copy_1 = self.test_rna.copy()
        del copy_1[-1]
        assert_equal(str(copy_1), 'AUG')

    def test_setitem(self):
        copy0 = self.test_rna.copy()
        copy0[0] = 'u'
        assert_equal(str(copy0), 'UUGC')
        copy1 = self.test_rna.copy()
        copy1[1] = 'a'
        assert_equal(str(copy1), 'AAGC')
        copy2 = self.test_rna.copy()
        copy2[2] = 'a'
        assert_equal(str(copy2), 'AUAC')
        copy3 = self.test_rna.copy()
        copy3[3] = 'a'
        assert_equal(str(copy3), 'AUGA')
        copy_1 = self.test_rna.copy()
        copy_1[-1] = 'a'
        assert_equal(str(copy_1), 'AUGA')

    def test_str(self):
        assert_equal(str(self.test_rna), 'AUGC')

    def test_len(self):
        assert_equal(len(self.test_rna), 4)

    def test_add(self):
        assert_equal(str((self.test_rna + self.test_rna)), 'AUGCAUGC')

    def test_radd(self):
        assert_equal(str(sum([self.test_rna, self.test_rna])), 'AUGCAUGC')

        def radd_800(seq):
            return 800 + seq

        assert_raises(TypeError, radd_800, self.test_rna)

    def test_mul(self):
        assert_equal(str((self.test_rna * 4)), 'AUGCAUGCAUGCAUGC')

        def mul_float(seq):
            return seq * 7.56

        assert_raises(TypeError, mul_float, self.test_rna)

    def test_eq(self):
        assert_true(self.test_rna == RNA('augc'))

    def test_ne(self):
        assert_true(self.test_rna != RNA('aagc'))

    def test_contains(self):
        assert_true(RNA('a') in self.test_rna)
        assert_true(RNA('u') in self.test_rna)
        assert_true(RNA('g') in self.test_rna)
        assert_true(RNA('c') in self.test_rna)
        assert_false(RNA('a') in RNA('ugc'))



# -*- coding: utf-8
'''Wrapper for NUPACK 3.0.'''
import multiprocessing
import numpy as np
import os
import re
import subprocess
import time
from coral.utils import tempdirs


class LambdaError(Exception):
    '''Raise if maximum states is exceeded (for \'distributions\' command).'''


class NUPACK(object):
    '''Run NUPACK functions on sequences.'''

    def __init__(self, nupack_home=None):
        '''
        :param nupack_home: NUPACK home dir. If the NUPACK commands aren't in
                            your path and the NUPACKHOME environment variable
                            isn't set, you can manually specify the NUPACK
                            directory here (the directory that contains bin/).
        :type nupack_home: str

        '''
        # Figure out where the NUPACK executables are
        if nupack_home is not None:
            self._nupack_home = nupack_home
        else:
            try:
                self._nupack_home = os.environ['NUPACKHOME']
            except KeyError:
                pfunc_paths = []
                for path in os.environ['PATH'].split(os.pathsep):
                    test = os.path.join(path, 'pfunc')
                    if os.path.isfile(test) and os.access(test, os.X_OK):
                        pfunc_paths.append(test)
                if not pfunc_paths:
                    raise IOError('NUPACK commands not found - see '
                                  'documentation')

        # Initialize empty temp dir location
        self._tempdir = ''

    @tempdirs.tempdir
    def pfunc(self, strand, temp=37.0, pseudo=False, material=None,
              dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the partition function for an ordered complex of strands.
        Runs the \'pfunc\' command.

        :param strand: Strand on which to run pfunc. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the free energy of the ordered complex
                  (float) and the partition function (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command
        stdout = self._run('pfunc', cmd_args, [str(strand)]).split('\n')

        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def pfunc_multi(self, strands, permutation=None, temp=37.0, pseudo=False,
                    material=None, dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the partition function for an ordered complex of strands.
        Runs the \'pfunc\' command.

        :param strands: List of strands to use as inputs to pfunc -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the free energy of the ordered complex
                  (float) and the partition function (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        stdout = self._run('pfunc', cmd_args, lines).split('\n')

        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def pairs(self, strand, cutoff=0.001, temp=37.0, pseudo=False,
              material=None, dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the pair probabilities for an ordered complex of strands.
        Runs the \'pairs\' command.

        :param strand: Strand on which to run pairs. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param cutoff: Only probabilities above this cutoff appear in the
                       output.
        :type cutoff: float
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The probability matrix, where the (i, j)th entry
                  is the probability that base i is bound to base j. The matrix
                  is augmented (it's N+1 by N+1, where N is the number of bases
                  in the sequence) with an (N+1)th column containing the
                  probability that each base is unpaired.
        :rtype: numpy.array

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand)]
        self._run('pairs', cmd_args, lines)

        # Read the output from file
        ppairs = self._read_tempfile('pairs.ppairs')
        data = re.search('\n\n\d*\n(.*)', ppairs, flags=re.DOTALL).group(1)
        N = len(strand)
        data_lines = [line.split('\t') for line in data.split('\n') if line]
        prob_matrix = self._pairs_to_np(data_lines, N)

        return prob_matrix

    @tempdirs.tempdir
    def pairs_multi(self, strands, cutoff=0.001, permutation=None, temp=37.0,
                    pseudo=False, material=None, dangles='some', sodium=1.0,
                    magnesium=0.0):
        '''Compute the pair probabilities for an ordered complex of strands.
        Runs the \'pairs\' command.

        :param strands: List of strands to use as inputs to pairs -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param cutoff: Only probabilities above this cutoff appear in the
                       output.
        :type cutoff: float
        :returns: Two probability matrices: The probability matrix as in the
                  pairs method (but with a dimension equal to the sum of the
                  lengths of the sequences in the permutation), and a similar
                  probability matrix where multiple strands of the same species
                  are considered to be indistinguishable.
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        self._run('pairs', cmd_args, lines)

        # Read the output from file
        N = sum([len(s) for s in strands])
        matrices = []
        for mat_type in ['ppairs', 'epairs']:
            data = self._read_tempfile('pairs.' + mat_type)
            probs = re.search('\n\n\d*\n(.*)', data, flags=re.DOTALL).group(1)
            lines = probs.split('\n')
            # Remove the last line (empty)
            lines.pop()
            pairlist = [line.split('\t') for line in lines]
            prob_matrix = self._pairs_to_np(pairlist, N)
            matrices.append(prob_matrix)

        return matrices

    @tempdirs.tempdir
    def mfe(self, strand, degenerate=False, temp=37.0, pseudo=False,
            material=None, dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the MFE for an ordered complex of strands. Runs the \'mfe\'
        command.

        :param strand: Strand on which to run mfe. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param degenerate: Setting to True will result in returning a list of
                           dictionaries associated with structures having the
                           same, minimal MFE value.
        :type degenerate: bool
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A dictionary with keys for 'mfe' (a float), 'dotparens'
                  (dot-parens notation of the MFE structure), and 'pairlist'
                  (a pair list notation of the MFE structure). Note that the
                  pair list will be an empty list if the MFE is unstructured.
        :rtype: dict

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)
        if degenerate:
            cmd_args.append('-degenerate')

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand)]
        self._run('mfe', cmd_args, lines)

        # Read the output from file
        structures = self._process_mfe(self._read_tempfile('mfe.mfe'))

        if degenerate:
            return structures
        else:
            return structures[0]

    @tempdirs.tempdir
    def mfe_multi(self, strands, permutation=None, degenerate=False, temp=37.0,
                  pseudo=False, material=None, dangles='some', sodium=1.0,
                  magnesium=0.0):
        '''Compute the MFE for an ordered complex of strands. Runs the \'mfe\'
        command.

        :param strands: Strands on which to run mfe. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :param degenerate: Setting to True will result in returning a list of
                           dictionaries associated with structures having the
                           same, minimal MFE value.
        :type degenerate: bool
        :returns: A dictionary with keys for 'mfe' (a float), 'dotparens'
                  (dot-parens notation of the MFE structure), and 'pairlist'
                  (a pair list notation of the MFE structure). Note that the
                  pair list will be an empty list if the MFE is unstructured.
        :rtype: dict

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        if degenerate:
            cmd_args.append('-degenerate')

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        self._run('mfe', cmd_args, lines)

        # Read the output from file
        structures = self._process_mfe(self._read_tempfile('mfe.mfe'))

        if degenerate:
            return structures
        else:
            return structures[0]

    @tempdirs.tempdir
    def subopt(self, strand, gap, temp=37.0, pseudo=False, material=None,
               dangles='some', sodium=1.0, magnesium=0.0):
        '''Compute the suboptimal structures within a defined energy gap of the
        MFE. Runs the \'subopt\' command.

        :param strand: Strand on which to run subopt. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param gap: Energy gap within to restrict results, e.g. 0.1.
        :type gap: float
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A list of dictionaries of the type returned by .mfe().
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand), str(gap)]
        self._run('subopt', cmd_args, lines)

        # Read the output from file
        structures = self._process_mfe(self._read_tempfile('subopt.subopt'))

        return structures

    @tempdirs.tempdir
    def subopt_multi(self, strands, gap, permutation=None, temp=37.0,
                     pseudo=False, material=None, dangles='some', sodium=1.0,
                     magnesium=0.0):
        '''Compute the suboptimal structures within a defined energy gap of the
        MFE for an ordered permutation of strands. Runs the \'subopt\' command.

        :param strands: Strands on which to run subopt. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param gap: Energy gap within to restrict results, e.g. 0.1.
        :type gap: float
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A list of dictionaries of the type returned by .mfe().
        :rtype: list

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)

        # Set up the input file and run the command. Note: no STDOUT
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(str(gap))
        self._run('subopt', cmd_args, lines)

        # Read the output from file
        structures = self._process_mfe(self._read_tempfile('subopt.subopt'))

        return structures

    @tempdirs.tempdir
    def count(self, strand, pseudo=False):
        '''Enumerates the total number of secondary structures over the
        structural ensemble Ω(π). Runs the \'count\' command.

        :param strand: Strand on which to run count. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: list
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :returns: The count of the number of structures in the structural
                  ensemble.
        :rtype: int

        '''
        # Set up command flags
        if pseudo:
            cmd_args = ['-pseudo']
        else:
            cmd_args = []

        # Set up the input file and run the command
        stdout = self._run('count', cmd_args, [str(strand)]).split('\n')

        # Return the count
        return int(float(stdout[-2]))

    @tempdirs.tempdir
    def count_multi(self, strands, permutation=None, pseudo=False):
        '''Enumerates the total number of secondary structures over the
        structural ensemble Ω(π) with an ordered permutation of strands. Runs
        the \'count\' command.

        :param strands: List of strands to use as inputs to count -multi.
        :type strands: list
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :returns: Dictionary with the following key:value pairs: 'energy':
                  free energy, 'pfunc': partition function.
        :rtype: dict

        '''
        # Set up command flags
        cmd_args = ['-multi']
        if pseudo:
            cmd_args.append('-pseudo')

        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        stdout = self._run('count', cmd_args, lines).split('\n')

        return int(float(stdout[-2]))

    @tempdirs.tempdir
    def energy(self, strand, dotparens, temp=37.0, pseudo=False, material=None,
               dangles='some', sodium=1.0, magnesium=0.0):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strand: Strand on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command. Note: no STDOUT
        lines = [str(strand), dotparens]
        stdout = self._run('energy', cmd_args, lines).split('\n')

        # Return the energy
        return float(stdout[-2])

    @tempdirs.tempdir
    def energy_multi(self, strands, dotparens, permutation=None, temp=37.0,
                     pseudo=False, material=None, dangles='some', sodium=1.0,
                     magnesium=0.0):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('energy', cmd_args, lines).split('\n')

        return float(stdout[-2])

    @tempdirs.tempdir
    def prob(self, strand, dotparens, temp=37.0, pseudo=False, material=None,
             dangles='some', sodium=1.0, magnesium=0.0):
        '''Calculate the equilibrium probability of a given secondary
        structure. Runs the \'prob\' command.

        :param strand: Strand on which to run prob. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The equilibrium probability of the sequence with the
                  specified secondary structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)

        # Set up the input file and run the command.
        lines = [str(strand), dotparens]
        stdout = self._run('prob', cmd_args, lines).split('\n')

        # Return the probabilities
        return float(stdout[-2])

    @tempdirs.tempdir
    def prob_multi(self, strands, dotparens, permutation=None, temp=37.0,
                   pseudo=False, material=None, dangles='some', sodium=1.0,
                   magnesium=0.0):
        '''Calculate the equilibrium probability of a given secondary
        structure of a complex of sequences in a given circular permutation.
        Runs the \'prob\' command.

        :param strands: Strands on which to run prob. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: The free energy of the sequence with the specified secondary
                  structure.
        :rtype: float

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('prob', cmd_args, lines).split('\n')

        return float(stdout[-2])

    @tempdirs.tempdir
    def defect(self, strand, dotparens, mfe=False, temp=37.0, pseudo=False,
               material=None, dangles='some', sodium=1.0, magnesium=0.0):
        '''Calculate the ensemble defect for a given sequence and secondary
        structure. From the documentation, the ensemble defect is defined as
        \'the number of incorrectly paired nucleotides at equilibrium evaluated
        over the ensemble of the ordered complex.\' Runs the \'defect\'
        command.

        :param strand: Strand on which to run defect. Strands must be either
                       coral.DNA or coral.RNA).
        :type strand: coral.DNA or coral.RNA
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param mfe: Return the MFE defect (Zadeh et al., 2010) instead of the
                    ensemble defect.
        :type mfe: bool
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the ensemble defect (float) and normalized
                  ensemble defect (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strand, material)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)
        if mfe:
            cmd_args.append('-mfe')

        # Set up the input file and run the command.
        lines = [str(strand), dotparens]
        stdout = self._run('defect', cmd_args, lines).split('\n')

        # Return the defect [ensemble defect, ensemble defect]
        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def defect_multi(self, strands, dotparens, permutation=None, mfe=False,
                     temp=37.0, pseudo=False, material=None, dangles='some',
                     sodium=1.0, magnesium=0.0):
        '''Calculate the free energy of a given sequence structure. Runs the
        \'energy\' command.

        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list
        :param dotparens: The structure in dotparens notation.
        :type dotparens: str
        :param permutation: The circular permutation of strands to test in
                            complex. e.g. to test in the order that was input
                            for 4 strands, the permutation would be [1,2,3,4].
                            If set to None, defaults to the order of the
                            input strands.
        :type permutation: list
        :param mfe: Return the MFE defect (Zadeh et al., 2010) instead of the
                    ensemble defect.
        :type mfe: bool
        :param temp: Temperature setting for the computation. Negative values
                     are not allowed.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A 2-tuple of the ensemble defect (float) and normalized
                  ensemble defect (float).
        :rtype: tuple

        '''
        # Set the material (will be used to set command material flag)
        material = self._set_material(strands, material, multi=True)

        # Set up command flags
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=True)
        if mfe:
            cmd_args.append('-mfe')

        # Set up the input file and run the command
        if permutation is None:
            permutation = range(1, len(strands) + 1)
        lines = self._multi_lines(strands, permutation)
        lines.append(dotparens)
        stdout = self._run('defect', cmd_args, lines).split('\n')

        # Return the defect [ensemble defect, ensemble defect]
        return (float(stdout[-3]), float(stdout[-2]))

    @tempdirs.tempdir
    def complexes(self, strands, max_size, ordered=False, pairs=False,
                  mfe=False, cutoff=0.001, degenerate=False, temp=37.0,
                  pseudo=False, material=None, dangles='some', sodium=1.0,
                  magnesium=0.0):
        '''
        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list of coral.DNA or coral.RNA
        :param max_size: Maximum complex size to consider (maximum number of
                         strand species in complex).
        :type max_size: int
        :param ordered: Consider distinct ordered complexes - all distinct
                        circular permutations of each complex.
        :type ordered: bool
        :param pairs: Calculate base-pairing observables as with .pairs().
        :type pairs: bool
        :param cutoff: A setting when pairs is set to True - only probabilities
                       above this threshold will be returned.
        :type cutoff: float
        :param degenerate: Applies only when \'mfe\' is set to True. If
                           set to True, the 'mfe' value associated with each
                           complex will be a list of degenerate MFEs (as in
                           the case of .mfe()).
        :type degenerate: bool
        :param temp: Temperature.
        :type temp: float
        :param pseudo: Enable pseudoknots.
        :type pseudo: bool
        :param material: The material setting to use in the computation. If set
                         to None (the default), the material type is inferred
                         from the strands. Other settings available: 'dna' for
                         DNA parameters, 'rna' for RNA (1995) parameters, and
                         'rna1999' for the RNA 1999 parameters.
        :type material: str
        :param dangles: How to treat dangles in the computation. From the
                        user guide: For \'none\': Dangle energies are ignored.
                        For \'some\': \'A dangle energy is incorporated for
                        each unpaired base flanking a duplex\'. For 'all': all
                        dangle energy is considered.
        :type dangles: str
        :param sodium: Sodium concentration in solution (molar), only applies
                       to DNA.
        :type sodium: float
        :param magnesium: Magnesium concentration in solution (molar), only
                          applies to DNA>
        :type magnesium: float
        :returns: A list of dictionaries containing at least 'energy',
                  'complex', and 'strands' keys. If 'ordered' is True, the
                  different possible ordered permutations of complexes are
                  considered. In addition, with 'ordered' set to True, an
                  additional 'order' key describing the exact order of strands
                  and a 'permutation' index (integer) are added. If 'pairs' is
                  True, there is an additional 'epairs' key containing the
                  base-pairing expectation values. If 'mfe' is selected, 'mfe,
                  'dotparens', and 'pairlist' keys in the same as .mfe(). In
                  addition, 'mfe' sets the -ordered flag, so the same keys as
                  when 'ordered' is set to True are added.
        :rtype: float

        '''
        # TODO: Consider returning a pandas dataframe in this (and other)
        # situations to make sorting/selection between results easier.
        material = self._set_material(strands, material, multi=True)
        cmd_args = self._prep_cmd_args(temp, dangles, material, pseudo, sodium,
                                       magnesium, multi=False)
        cmd_args.append('-quiet')
        if mfe:
            cmd_args.append('-mfe')
            ordered = True
            if degenerate:
                cmd_args.append('-degenerate')
        if ordered:
            cmd_args.append('-ordered')
        if pairs:
            cmd_args.append('-pairs')
            cmd_args.append('-cutoff')
            cmd_args.append(cutoff)

        dim = sum([len(s) for s in strands])
        nstrands = len(strands)
        # Set up the input file and run the command
        lines = self._multi_lines(strands, [max_size])
        self._run('complexes', cmd_args, lines)

        # Read the output from file(s)
        if ordered:
            ocx_lines = self._read_tempfile('complexes.ocx').split('\n')
            # Process each lines
            output = []
            for line in ocx_lines:
                if line and not line.startswith('%'):
                    data = line.split('\t')
                    energy = float(data[-1])
                    complexes = [int(d) for d in data[2:2 + nstrands]]
                    permutation = int(data[1])
                    output.append({'energy': energy, 'complex': complexes,
                                   'permutation': permutation})

            key_lines = self._read_tempfile('complexes.ocx-key').split('\n')

            data_lines = [l for l in key_lines if not l.startswith('%')]
            data_lines.pop()
            for i, line in enumerate(data_lines):
                data = line.split('\t')
                keys = [int(d) for d in data[2:-1]]
                output[i]['order'] = keys

            if pairs:
                epairs_data = self._read_tempfile('complexes.ocx-epairs')
                pairslist = self._process_epairs(epairs_data)
                for i, pairs in enumerate(pairslist):
                    output[i]['epairs'] = self._pairs_to_np(pairs, dim)
                # TODO: add ocx-ppairs as well

            if mfe:
                mfe_data = self._read_tempfile('complexes.ocx-mfe')
                if degenerate:
                    return NotImplementedError('Not implemented for complexes')
                else:
                    mfe_output = self._process_mfe(mfe_data, complexes=True)
                    for i, mfedat in enumerate(mfe_output):
                        output[i]['mfe'] = mfedat['mfe']
                        output[i]['dotparens'] = mfedat['dotparens']
                        output[i]['pairlist'] = mfedat['pairlist']
        else:
            cx_lines = self._read_tempfile('complexes.cx').split('\n')
            # Remove empty last line
            cx_lines.pop()
            output = []
            for line in cx_lines:
                if not line.startswith('%'):
                    data = line.split('\t')
                    energy = float(data[-1])
                    complexes = [int(d) for d in data[1:1 + len(strands)]]
                    output.append({'energy': energy, 'complex': complexes})

            if pairs:
                # Process epairs
                epairs_data = self._read_tempfile('complexes.cx-epairs')
                pairslist = self._process_epairs(epairs_data)
                for i, pairs in enumerate(pairslist):
                    proba_mat = self._pairs_to_np(pairs, dim)
                    output[i]['epairs'] = proba_mat

        # Add strands (for downstream concentrations)
        for cx in output:
            cx['strands'] = [s.copy() for s in strands]

        return output

    @tempdirs.tempdir
    def complexes_timeonly(self, strands, max_size):
        '''Estimate the amount of time it will take to calculate all the
        partition functions for each circular permutation - estimate the time
        the actual \'complexes\' command will take to run.

        :param strands: Strands on which to run energy. Strands must be either
                       coral.DNA or coral.RNA).
        :type strands: list of coral.DNA or coral.RNA
        :param max_size: Maximum complex size to consider (maximum number of
                         strand species in complex).
        :type max_size: int
        :returns: The estimated time to run complexes' partition functions, in
                  seconds.
        :rtype: float

        '''
        cmd_args = ['-quiet', '-timeonly']
        lines = self._multi_lines(strands, [max_size])
        stdout = self._run('complexes', cmd_args, lines)
        return float(re.search('calculation\: (.*) seconds', stdout).group(1))

    @tempdirs.tempdir
    def concentrations(self, complexes, concs, ordered=False, pairs=False,
                       cutoff=0.001, temp=37.0):
        '''
        :param complexes: A list of the type returned by the complexes()
                          method.
        :type complexes: list
        :param concs: The concentration(s) of each strand species in the
                      initial complex. If they are all the same, a single
                      float can be used here.
        :type concs: list of floats or float
        :param ordered: Consider distinct ordered complexes - all distinct
                        circular permutations of each complex.
        :type ordered: bool
        :param pairs: Calculate base-pairing observables as with .pairs().
        :type pairs: bool
        :param cutoff: A setting when pairs is set to True - only probabilities
                       above this threshold will be returned.
        :type cutoff: float
        :param temp: Temperature in C.
        :type temp: float
        :returns: A list of dictionaries containing (at least) a
                  'concentrations' key. If 'pairs' is True, an 'fpairs' key
                  is added.
        :rtype: list

        '''
        # Check inputs
        nstrands = len(complexes[0]['strands'])
        try:
            if len(concs) != nstrands:
                raise ValueError('concs argument not same length as strands.')
        except TypeError:
            concs = [concs for i in range(len(complexes['strands']))]

        # Set up command-line arguments
        cmd_args = ['-quiet']
        if ordered:
            cmd_args.append('-ordered')

        # Write .con file
        with open(os.path.join(self._tempdir, 'concentrations.con')) as f:
            f.writelines(concs)

        # Write .cx or .ocx file
        header = ['%t Number of strands: {}'.format(nstrands),
                  '%\tid\tsequence']
        for i, strand in enumerate(complexes['strands']):
            header.append('%\t{}\t{}'.format(i + 1, strand))
        header.append('%\tT = {}'.format(temp))
        body = []
        for i, cx in enumerate(complexes):
            permutation = '\t'.join(complexes['complex'])
            line = '{}\t{}\t{}'.format(i + 1, permutation, complexes['energy'])
            body.append(line)

        if ordered:
            cxfile = os.path.join(self._tempdir, 'concentrations.ocx')
        else:
            cxfile = os.path.join(self._tempdir, 'concentrations.cx')

        with open(cxfile) as f:
            f.writelines(header + body)

        # Run 'concentrations'
        self._run('concentrations', cmd_args, None)

        # Parse the .eq (concentrations) file
        eq_lines = self._read_tempfile('concentrations.eq').split('\n')
        tsv_lines = [l for l in eq_lines if not l.startswith('%')]
        output = []
        for i, line in enumerate(tsv_lines):
            # It's a TSV
            data = line.split('\t')
            # Column 0 is an index
            # Columns 1-nstrands is the complex
            cx = [int(c) for c in data[1:nstrands]]
            # Column nstrands + 1 is the complex energy
            # Column nstrands + 2 is the equilibrium concentration
            eq = float(data[nstrands + 2])
            output[i] = {'complex': cx, 'concentration': eq}

        if pairs:
            # Read the .fpairs file
            pairs = self._read_tempfile('concentrations.fpairs')
            pairs_tsv = [l for l in pairs.split('\n') if not l.startswith('%')]
            # Remove first line (n complexes)
            dim = int(pairs_tsv.pop(0))
            pprob = [[int(p[0]), int(p[1]), float(p[2])] for p in pairs_tsv]
            # Convert to augmented numpy matrix
            fpairs_mat = self.pairs_to_np(pprob, dim)
            for i, out in enumerate(output):
                output[i]['fpairs'] = fpairs_mat

        return output

    @tempdirs.tempdir
    def distributions(self, complexes, counts, volume, maxstates=1e7,
                      ordered=False, temp=37.0):
        '''Runs the \'distributions\' NUPACK command. Note: this is intended
        for a relatively small number of species (on the order of ~20
        total strands for complex size ~14).

        :param complexes: A list of the type returned by the complexes()
                          method.
        :type complexes: list
        :param counts: A list of the exact number of molecules of each initial
                       species (the strands in the complexes command).
        :type counts: list of ints
        :param volume: The volume, in liters, of the container.
        :type volume: float
        :param maxstates: Maximum number of states to be enumerated, needed
                          as allowing too many states can lead to a segfault.
                          In NUPACK, this is referred to as lambda.
        :type maxstates: float
        :param ordered: Consider distinct ordered complexes - all distinct
                        circular permutations of each complex.
        :type ordered: bool
        :param temp: Temperature in C.
        :type temp: float
        :returns: A list of dictionaries containing (at least) a 'complexes'
                  key for the unique complex, an 'ev' key for the expected
                  value of the complex population and a 'probcols' list
                  indicating the probability that a given complex has
                  population 0, 1, ... max(pop) at equilibrium.
        :rtype: list
        :raises: LambdaError if maxstates is exceeded.

        '''
        # Check inputs
        nstrands = len(complexes[0]['strands'])
        if len(counts) != nstrands:
            raise ValueError('counts argument not same length as strands.')

        # Set up command-line arguments
        cmd_args = []
        if ordered:
            cmd_args.append('-ordered')

        # Write .count file
        countpath = os.path.join(self._tempdir, 'distributions.count')
        with open(countpath, 'w') as f:
            f.writelines([str(c) for c in counts] + [str(volume)])

        # Write .cx or .ocx file
        header = ['%t Number of strands: {}'.format(nstrands),
                  '%\tid\tsequence']
        for i, strand in enumerate(complexes['strands']):
            header.append('%\t{}\t{}'.format(i + 1, strand))
        header.append('%\tT = {}'.format(temp))
        body = []
        for i, cx in enumerate(complexes):
            permutation = '\t'.join(complexes['complex'])
            line = '{}\t{}\t{}'.format(i + 1, permutation, complexes['energy'])
            body.append(line)

        if ordered:
            cxfile = os.path.join(self._tempdir, 'distributions.ocx')
        else:
            cxfile = os.path.join(self._tempdir, 'distributions.cx')

        with open(cxfile) as f:
            f.writelines(header + body)

        # Run 'distributions'
        stdout = self._run('distributions', cmd_args, None)

        # Parse STDOUT
        stdout_lines = stdout.split('\n')
        if stdout_lines[0].startswith('Exceeded maximum number'):
            raise LambdaError('Exceeded maxstates combinations.')

        # pop_search = re.search('There are (*) pop', stdout_lines[0]).group(1)
        # populations = int(pop_search)
        # kT_search = re.search('of the box: (*) kT', stdout_lines[1]).group(1)
        # kT = float(kT_search)

        # Parse .dist file (comments header + TSV)
        dist_lines = self._read_tempfile('distributions.dist').split('\n')
        tsv_lines = [l for l in dist_lines if not l.startswith('%')]
        tsv_lines.pop()
        output = []
        for i, line in enumerate(tsv_lines):
            data = line.split('\t')
            # Column 0 is an index
            # Columns 1-nstrands are complexes
            cx = [int(d) for d in data[1:nstrands]]
            # Column nstrands + 1 is expected value of complex
            ev = float(data[nstrands + 1])
            # Columns nstrands + 2 and on are probability columns
            probcols = [float(d) for d in data[nstrands + 2:]]
            output[i]['complex'] = cx
            output[i]['ev'] = ev
            output[i]['probcols'] = probcols

        return output

    # Helper methods for preparing command input files
    def _multi_lines(self, strands, permutation):
        '''Prepares lines to write to file for pfunc command input.

        :param strand: Strand input (cr.DNA or cr.RNA).
        :type strand: cr.DNA or cr.DNA
        :param permutation: Permutation (e.g. [1, 2, 3, 4]) of the type used
                            by pfunc_multi.
        :type permutation: list
        '''
        lines = []
        # Write the total number of distinct strands
        lines.append(str(len(strands)))
        # Write the distinct strands
        lines += [str(strand) for strand in strands]
        # Write the permutation
        lines.append(' '.join(str(p) for p in permutation))

        return lines

    # Helper methods for processing output files
    def _read_tempfile(self, filename):
        '''Read in and return file that's in the tempdir.

        :param filename: Name of the file to read.
        :type filename: str

        '''
        with open(os.path.join(self._tempdir, filename)) as f:
            return f.read()

    def _pairs_to_np(self, pairlist, dim):
        '''Given a set of pair probability lines, construct a numpy array.

        :param pairlist: a list of pair probability triples
        :type pairlist: list
        :returns: An upper triangular matrix of pair probabilities augmented
                  with one extra column that represents the unpaired
                  probabilities.
        :rtype: numpy.array

        '''
        mat = np.zeros((dim, dim + 1))
        for line in pairlist:
            i = int(line[0]) - 1
            j = int(line[1]) - 1
            prob = float(line[2])
            mat[i, j] = prob
        return mat

    def _process_mfe(self, data, complexes=False):
        # Parse the output data
        # Find the text in between the large comment lines
        commentline = '\n% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %\n'
        # Find everything between two commentlines
        groups = data.split(commentline)
        # Everything before the comment line is notes about the command,
        # the last part is just a newline
        groups.pop(0)
        groups.pop()
        # The remainder is data
        output = []
        # Skip every other one (every 2nd match is empty lines)
        for group in groups[::2]:
            lines = group.split('\n')
            # Line 1 is the strand number (ignored)
            if complexes:
                lines.pop(0)
            lines.pop(0)
            # Line 2 is the MFE
            mfe = float(lines.pop(0))
            # Line 3 is the dot-bracket structure
            dotparens = lines.pop(0)
            # If there are any more lines, they are a pair list format
            # structure
            pairlist = []
            for line in lines:
                pair = line.split('\t')
                pairlist.append([int(pair[0]) - 1, int(pair[1]) - 1])
            output.append({'mfe': mfe, 'dotparens': dotparens,
                           'pairlist': pairlist})

        return output

    def _process_epairs(self, filedata):
        commentline = '\n% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %\n'
        groups = filedata.split(commentline)
        groups.pop(0)
        groups.pop()
        output = []
        for group in groups[::2]:
            lines = group.split('\n')
            lines.pop(0)
            lines.pop(0)
            output.append([line.split('\t') for line in lines])
        return output

    # Helper methods for repetitive tasks
    def _set_material(self, strand_input, material, multi=False):
        if multi:
            if len(set([s.material for s in strand_input])) > 1:
                raise ValueError('Inputs must all be coral.DNA or all '
                                 'coral.RNA')
            if material is None:
                return strand_input[0].material
            else:
                return material
        else:
            if material is None:
                return strand_input.material
            else:
                return material

    def _prep_cmd_args(self, temp, dangles, material, pseudo, sodium,
                       magnesium, multi=False):
        cmd_args = []
        cmd_args += ['-T', temp]
        cmd_args += ['-dangles', dangles]
        cmd_args += ['-material', material]
        cmd_args += ['-sodium', sodium]
        cmd_args += ['-magnesium', magnesium]
        if pseudo:
            cmd_args.append('-pseudo')
        if multi:
            cmd_args.append('-multi')

        return cmd_args

    def _run(self, command, cmd_args, lines):
        prefix = command
        path = os.path.join(self._tempdir, '{}.in'.format(prefix))
        with open(path, 'w') as f:
            f.write('\n'.join(lines))

        arguments = [os.path.join(self._nupack_home, 'bin', command)]
        arguments += cmd_args
        arguments.append(prefix)

        arguments = [str(x) for x in arguments]
        process = subprocess.Popen(arguments, stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT, cwd=self._tempdir)
        output = process.communicate()[0]
        return output


def nupack_multi(seqs, material, cmd, arguments, report=True):
    '''Split Nupack commands over processors.

    :param inputs: List of sequences, same format as for coral.analysis.Nupack.
    :type inpus: list
    :param material: Input material: 'dna' or 'rna'.
    :type material: str
    :param cmd: Command: 'mfe', 'pairs', 'complexes', or 'concentrations'.
    :type cmd: str
    :param arguments: Arguments for the command.
    :type arguments: str
    :returns: A list of the same return value you would get from `cmd`.
    :rtype: list

    '''
    nupack_pool = multiprocessing.Pool()
    try:
        args = [{'seq': seq,
                 'cmd': cmd,
                 'material': material,
                 'arguments': arguments} for seq in seqs]
        nupack_iterator = nupack_pool.imap(run_nupack, args)
        total = len(seqs)
        msg = ' calculations complete.'
        passed = 4
        while report:
            completed = nupack_iterator._index
            if (completed == total):
                break
            else:
                if passed >= 4:
                    print '({0}/{1}) '.format(completed, total) + msg
                    passed = 0
                passed += 1
                time.sleep(1)
        multi_output = [x for x in nupack_iterator]
        nupack_pool.close()
        nupack_pool.join()
    except KeyboardInterrupt:
        nupack_pool.terminate()
        nupack_pool.close()
        raise KeyboardInterrupt

    return multi_output


def run_nupack(kwargs):
    '''Run picklable Nupack command.

    :param kwargs: keyword arguments to pass to Nupack as well as 'cmd'.
    :returns: Variable - whatever `cmd` returns.

    '''
    run = NUPACK(kwargs['seq'])
    output = getattr(run, kwargs['cmd'])(**kwargs['arguments'])
    return output



'''
Tests for RandomCodons class of analysis module.

'''

from nose.tools import assert_equal, assert_not_equal, assert_raises
from coral import design, reaction, RNA


def test_randomcodons():
    '''
    This test is pretty basic right now - not sure how much checking
    can be done for a random DNA base generator.

    '''

    reference_seq = RNA('AUGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAUAG')
    reference_peptide = reaction.translate(reference_seq)
    output = design.random_codons(reference_peptide)
    output_peptide = reaction.translate(reference_seq)

    assert_equal(len(output), len(reference_seq) - 3)
    assert_equal(reference_peptide, output_peptide)
    assert_not_equal(reference_seq, output)

    # Setting too high a threshold should raise ValueError
    assert_raises(ValueError, design.random_codons, reference_peptide,
                  frequency_cutoff=1.5)

    # Weighted should work
    w_output = design.random_codons(reference_peptide, weighted=True)
    w_output_peptide = reaction.translate(reference_seq)

    assert_equal(len(w_output), len(reference_seq) - 3)
    assert_equal(reference_peptide, w_output_peptide)
    assert_not_equal(reference_seq, w_output)



'''Tests primer design module.'''
from nose.tools import assert_equals, assert_not_equal, assert_raises
from coral import design, DNA


def test_primer():
    '''Test primer function.'''
    seq = 'ATGGTGAGCAAGGGCGAGGAGCTGTTCACCGGGGTGGTGCCCATCCTGGTCGAGCTGGACGGC' + \
          'GACGTAAACGGCCACAAGTTCAGCGTGTCCGGCGAGGGCGAGGGCGATGCCACCTACGGCAAG' + \
          'CTGACCCTGAAGTTCATCTGCACCACCGGCAAGCTGCCCGTGCCCTGGCCCACCCTCGTGACC' + \
          'ACCTTCGGCTACGGCCTGCAGTGCTTCGCCCGCTACCCCGACCACATGAAGCAGCACGACTTC' + \
          'TTCAAGTCCGCCATGCCCGAAGGCTACGTCCAGGAGCGCACCATCTTCTTCAAGGACGACGGC' + \
          'AACTACAAGACCCGCGCCGAGGTGAAGTTCGAGGGCGACACCCTGGTGAACCGCATCGAGCTG' + \
          'AAGGGCATCGACTTCAAGGAGGACGGCAACATCCTGGGGCACAAGCTGGAGTACAACTACAAC' + \
          'AGCCACAACGTCTATATCATGGCCGACAAGCAGAAGAACGGCATCAAGGTGAACTTCAAGATC' + \
          'CGCCACAACATCGAGGACGGCAGCGTGCAGCTCGCCGACCACTACCAGCAGAACACCCCCATC' + \
          'GGCGACGGCCCCGTGCTGCTGCCCGACAACCACTACCTGAGCTACCAGTCCGCCCTGAGCAAA' + \
          'GACCCCAACGAGAAGCGCGATCACATGGTCCTGCTGGAGTTCGTGACCGCCGCCGGGATCACT' + \
          'CTCGGCATGGACGAGCTGTACAAGTAA'
    dna_seq = DNA(seq)
    primer = design.primer(dna_seq, tm=72, min_len=10, tm_undershoot=1,
                           tm_overshoot=3, end_gc=False,
                           tm_parameters='cloning', overhang=None)
    assert_equals(str(primer), 'ATGGTGAGCAAGGGCGAGGAG')
    # Ensure that overhang is appropriately applied
    overhang_primer = design.primer(dna_seq, tm=72, min_len=10,
                                    tm_undershoot=1, tm_overshoot=3,
                                    end_gc=False,
                                    tm_parameters='cloning',
                                    overhang=DNA('GATCGATAT'))
    assert_equals(str(overhang_primer), 'GATCGATATATGGTGAGCAAGGGCGAGGAG')
    # If sequence is too short (too low of Tm), raise ValueError
    too_short = DNA('at')
    assert_raises(ValueError, design.primer, too_short, tm=72)
    # Should design different primers (sometimes) if ending on GC is preferred
    diff_template = DNA('GATCGATCGATACGATCGATATGCGATATGATCGATAT')
    nogc = design.primer(diff_template, tm=72, min_len=10,
                         tm_undershoot=1, tm_overshoot=3, end_gc=False,
                         tm_parameters='cloning', overhang=None)
    withgc = design.primer(diff_template, tm=72, min_len=10,
                           tm_undershoot=1, tm_overshoot=3,
                           end_gc=True, tm_parameters='cloning', overhang=None)
    assert_not_equal(nogc, withgc)
    # Should raise ValueError if it's impossible to create an end_gc primer
    end_at_template = DNA('ATGCGATACGATACGCGATATGATATATatatatat' +
                          'ATAAaaaaaaaaaattttttttTTTTTTTTTTTTTT' +
                          'TTTTTTTTTT')
    assert_raises(ValueError, design.primer, end_at_template,
                  end_gc=True, tm=72)


def test_primers():
    '''Test primers function.'''
    seq = 'ATGGTGAGCAAGGGCGAGGAGCTGTTCACCGGGGTGGTGCCCATCCTGGTCGAGCTGGACGGC' + \
          'GACGTAAACGGCCACAAGTTCAGCGTGTCCGGCGAGGGCGAGGGCGATGCCACCTACGGCAAG' + \
          'CTGACCCTGAAGTTCATCTGCACCACCGGCAAGCTGCCCGTGCCCTGGCCCACCCTCGTGACC' + \
          'ACCTTCGGCTACGGCCTGCAGTGCTTCGCCCGCTACCCCGACCACATGAAGCAGCACGACTTC' + \
          'TTCAAGTCCGCCATGCCCGAAGGCTACGTCCAGGAGCGCACCATCTTCTTCAAGGACGACGGC' + \
          'AACTACAAGACCCGCGCCGAGGTGAAGTTCGAGGGCGACACCCTGGTGAACCGCATCGAGCTG' + \
          'AAGGGCATCGACTTCAAGGAGGACGGCAACATCCTGGGGCACAAGCTGGAGTACAACTACAAC' + \
          'AGCCACAACGTCTATATCATGGCCGACAAGCAGAAGAACGGCATCAAGGTGAACTTCAAGATC' + \
          'CGCCACAACATCGAGGACGGCAGCGTGCAGCTCGCCGACCACTACCAGCAGAACACCCCCATC' + \
          'GGCGACGGCCCCGTGCTGCTGCCCGACAACCACTACCTGAGCTACCAGTCCGCCCTGAGCAAA' + \
          'GACCCCAACGAGAAGCGCGATCACATGGTCCTGCTGGAGTTCGTGACCGCCGCCGGGATCACT' + \
          'CTCGGCATGGACGAGCTGTACAAGTAA'
    dna_seq = DNA(seq)
    primers_list = design.primers(dna_seq, tm=72, min_len=10,
                                  tm_undershoot=1, tm_overshoot=3,
                                  end_gc=False, tm_parameters='cloning',
                                  overhangs=None)
    primers = [str(x.primer()) for x in primers_list]
    assert_equals(primers, ['ATGGTGAGCAAGGGCGAGGAG',
                            'TTACTTGTACAGCTCGTCCATGCCG'])



'''
Tests for utils submodule of the analysis module.

'''

from nose.tools import assert_equal, assert_raises
from coral import analysis, DNA, RNA, Peptide


def test_utils():
    test_DNA = DNA('ATAGCGATACGAT')
    test_RNA = RNA('AUGCGAUAGCGAU')
    test_peptide = Peptide('msvkkkpvqg')
    test_str = 'msvkkkpvgq'

    assert_equal(analysis.utils.sequence_type(test_DNA), 'dna')
    assert_equal(analysis.utils.sequence_type(test_RNA), 'rna')
    assert_equal(analysis.utils.sequence_type(test_peptide), 'peptide')
    assert_raises(Exception, analysis.utils.sequence_type, test_str)



'''Tests for the DNA sequence class.'''
from coral import DNA, Feature, RestrictionSite
from nose.tools import assert_equal, assert_false, assert_true, assert_raises
from nose.tools import assert_not_equal


class TestDNA(object):
    '''Testing class for DNA'''
    def __init__(self):
        self.test_dna = DNA('atgc')

    def test_reverse_complement(self):
        assert_equal(str(self.test_dna.reverse_complement()), 'GCAT')

    def test_linearize(self):
        circ_dna = self.test_dna.circularize()
        assert_equal(str(circ_dna.linearize()), 'ATGC')
        assert_equal(str(circ_dna.linearize(0)), 'ATGC')
        assert_equal(str(circ_dna.linearize(1)), 'TGCA')
        assert_equal(str(circ_dna.linearize(2)), 'GCAT')
        assert_equal(str(circ_dna.linearize(3)), 'CATG')
        assert_equal(str(circ_dna.linearize(-1)), 'CATG')
        assert_raises(ValueError, circ_dna.linearize().linearize)

    def test_locate(self):
        assert_equal(self.test_dna.locate('a'), [[0], [2]])
        assert_equal(self.test_dna.locate('at'), [[0], [2]])
        assert_equal(self.test_dna.locate('gc'), [[2], [0]])
        assert_equal(self.test_dna.locate('atgg'), [[], []])
        # Circular DNA tests
        assert_equal(self.test_dna.circularize().locate('a'), [[0], [2]])
        assert_equal(self.test_dna.circularize().locate('at'), [[0], [2]])
        assert_equal(self.test_dna.circularize().locate('gc'), [[2], [0]])
        assert_equal(self.test_dna.circularize().locate('atgg'), [[], []])

    def test_copy(self):
        assert_equal(self.test_dna, self.test_dna.copy())

    def test_palindrome(self):
        palindromic_seq_even = DNA('ATGCGCAT')
        nonpalindromic_seq_even = DNA('ATGCGCAA')
        almost_palindrome_odd = DNA('ATGCCAT')
        assert_true(palindromic_seq_even.is_palindrome())
        assert_false(nonpalindromic_seq_even.is_palindrome())
        assert_false(almost_palindrome_odd.is_palindrome())

    def test_getitem(self):
        assert_equal(str(self.test_dna[0]), 'A')
        assert_equal(str(self.test_dna[1]), 'T')
        assert_equal(str(self.test_dna[2]), 'G')
        assert_equal(str(self.test_dna[3]), 'C')
        assert_equal(str(self.test_dna[-1]), 'C')

    def test_delitem(self):
        copy0 = self.test_dna.copy()
        del copy0[0]
        assert_equal(str(copy0), 'TGC')
        copy1 = self.test_dna.copy()
        del copy1[1]
        assert_equal(str(copy1), 'AGC')
        copy2 = self.test_dna.copy()
        del copy2[2]
        assert_equal(str(copy2), 'ATC')
        copy3 = self.test_dna.copy()
        del copy3[3]
        assert_equal(str(copy3), 'ATG')
        copy_1 = self.test_dna.copy()
        del copy_1[-1]
        assert_equal(str(copy_1), 'ATG')

    def test_setitem(self):
        copy0 = self.test_dna.copy()
        copy0[0] = 't'
        assert_equal(str(copy0), 'TTGC')
        copy1 = self.test_dna.copy()
        copy1[1] = 'a'
        assert_equal(str(copy1), 'AAGC')
        copy2 = self.test_dna.copy()
        copy2[2] = 'a'
        assert_equal(str(copy2), 'ATAC')
        copy3 = self.test_dna.copy()
        copy3[3] = 'a'
        assert_equal(str(copy3), 'ATGA')
        copy_1 = self.test_dna.copy()
        copy_1[-1] = 'a'
        assert_equal(str(copy_1), 'ATGA')

        def set_gap(seq):
            seq[2] = '-'
        assert_raises(ValueError, set_gap, self.test_dna)

    def test_repr(self):
        expected_repr = 'ATGC\nTACG'
        assert_equal(repr(self.test_dna), expected_repr)

        expected_circ_repr = 'ATGC\nTACG'
        assert_equal(repr(self.test_dna.circularize()), expected_circ_repr)

        repr_1 = 'ATGCATGCATGCATGCATGCATGCATGCATGCATGCATGC ... '
        repr_2 = 'ATGCATGCATGCATGCATGCATGCATGCATGCATGCATGC\n'
        repr_3 = 'TACGTACGTACGTACGTACGTACGTACGTACGTACGTACG ... '
        repr_4 = 'TACGTACGTACGTACGTACGTACGTACGTACGTACGTACG'
        expected_long_repr = repr_1 + repr_2 + repr_3 + repr_4
        assert_equal(repr(self.test_dna * 50), expected_long_repr)

    def test_str(self):
        assert_equal(str(self.test_dna), 'ATGC')

    def test_len(self):
        assert_equal(len(self.test_dna), 4)

    def test_add(self):
        assert_equal(str(self.test_dna + self.test_dna), 'ATGCATGC')

    def test_radd(self):
        assert_equal(str(sum([self.test_dna, self.test_dna])), 'ATGCATGC')

        def radd_800(seq):
            return 800 + seq

        assert_raises(TypeError, radd_800, self.test_dna)

    def test_mul(self):
        assert_equal(str(self.test_dna * 4), 'ATGCATGCATGCATGC')

        def mul_float(seq):
            return seq * 7.56

        assert_raises(TypeError, mul_float, self.test_dna)

        # TODO: reimplement this test using manual sequence input
        # def mul_incompatible(seq):
        #     return seq * 3

        # incompatible_seq = self.test_dna.copy()
        # incompatible_seq = incompatible_seq.five_resect(1)
        # incompatible_seq = incompatible_seq.reverse_complement()
        # incompatible_seq = incompatible_seq.five_resect(1)
        # incompatible_seq = incompatible_seq.reverse_complement()
        # assert_raises(Exception, mul_incompatible, incompatible_seq)

    def test_eq(self):
        assert_true(self.test_dna == DNA('atgc'))

    def test_ne(self):
        assert_true(self.test_dna != DNA('aagc'))

    def test_contains(self):
        assert_true('a' in self.test_dna)
        assert_true('t' in self.test_dna)
        assert_true('g' in self.test_dna)
        assert_true('c' in self.test_dna)
        assert_false('u' in self.test_dna)

    def test_flip(self):
        flipped = self.test_dna.flip()
        assert_equal(str(self.test_dna), flipped.bottom)
        assert_equal(self.test_dna.bottom, str(flipped))


class TestFeatures(object):
    '''Test features model using DNA object.'''
    def __init__(self):
        self.dna = DNA('ATGC') * 50
        self.apply_features()

    def apply_features(self):
        misc_feature = Feature('Misc Feature', 1, 20, 'misc_feature')
        misc_1_feature = Feature('Misc Feature', 1, 20, 'misc_feature',
                                 strand=1)
        coding_feature = Feature('Coding Feature', 21, 40, 'CDS')
        primer_feature = Feature('Primer Feature', 41, 60, 'primer_bind')
        promoter_feature = Feature('Promoter Feature', 61, 80, 'promoter')
        terminator_feature = Feature('Terminator Feature', 81, 100,
                                     'terminator')
        rbs_feature = Feature('RBS Feature', 101, 120, 'RBS')
        origin_feature = Feature('Origin Feature', 121, 140, 'rep_origin')
        utr3_feature = Feature("3'UTR Feature", 141, 160, "3'UTR")
        origin_feature2 = Feature('Origin Feature', 161, 180, 'rep_origin')

        input_features = [misc_feature, misc_1_feature, coding_feature,
                          primer_feature, promoter_feature, terminator_feature,
                          rbs_feature, origin_feature, utr3_feature,
                          origin_feature2]
        self.dna.features = input_features

    def test_good_features(self):
        for feature in self.dna.features:
            assert_true(feature.copy() in self.dna.features)

    def test_rev_comp(self):
        rev = self.dna.reverse_complement()
        for feature, rev_feature in zip(self.dna.features, rev.features):
            assert_not_equal(feature.strand, rev_feature.strand)
            assert_equal(len(self.dna) - feature.start, rev_feature.stop)
            assert_equal(len(self.dna) - feature.stop, rev_feature.start)

    def test_extract(self):
        test_utr3_feature = [feature for feature in self.dna.features if
                             feature.name == "3'UTR Feature"][0]
        extracted = self.dna.extract(test_utr3_feature)
        assert_equal(str(extracted), 'TGCATGCATGCATGCATGC')

    def test_excise(self):
        copy = self.dna.circularize()
        feature = copy.select_features('Coding Feature')[0]
        rotated = copy.rotate_to(feature.start)
        expected = rotated[feature.stop - feature.start:]

        backbone = copy.excise(feature)

        assert_equal(expected, backbone)

    def test_getitem(self):
        subsequence = self.dna[30:100]
        remaining_features = [Feature('Primer Feature', 11, 30, 'primer_bind'),
                              Feature('Promoter Feature', 31, 50, 'promoter'),
                              Feature('Terminator Feature', 51, 70,
                                      'terminator')]

        assert_equal(subsequence.features, remaining_features)
        assert_false(self.dna[10].features)
        new_seq = DNA('ATGC', features=[Feature('A', 0, 0, 'misc_feature')])
        assert_equal(new_seq[0].features[0],
                     Feature('A', 0, 0, 'misc_feature'))

    def test_ne(self):
        '''Test != operator'''
        assert_true(self.dna.features[0] != self.dna.features[4])
        assert_false(self.dna.features[0] != self.dna.features[0])


class TestRestrictionSite(object):
    '''Test RestrictionSite class.'''
    def __init__(self):
        self.ecorv = RestrictionSite(DNA('GATATC'), (3, 3), name='EcoRV')
        self.foki = RestrictionSite(DNA('GGATG'), (14, 18), name='FokI')

    def test_cuts_outside(self):
        '''Test cuts_outside method.'''
        assert_false(self.ecorv.cuts_outside())
        assert_true(self.foki.cuts_outside())

    def test_len(self):
        '''Test len function.'''
        assert_equal(len(self.ecorv), 6)
        assert_equal(len(self.foki), 5)



'''
Tests utils submodule of reaction module.

'''

from nose.tools import assert_equal, assert_raises
from coral import reaction, DNA


def test_convert_sequence():
    '''Tests DNA translation function.'''

    seq = 'ATGGTGAGCAAGGGCGAGGAGCTGTTCACCGGGGTGGTGCCCATCCTGGTCGAGCTGGACGGC' + \
          'GACGTAAACGGCCACAAGTTCAGCGTGTCCGGCGAGGGCGAGGGCGATGCCACCTACGGCAAG' + \
          'CTGACCCTGAAGTTCATCTGCACCACCGGCAAGCTGCCCGTGCCCTGGCCCACCCTCGTGACC' + \
          'ACCTTCGGCTACGGCCTGCAGTGCTTCGCCCGCTACCCCGACCACATGAAGCAGCACGACTTC' + \
          'TTCAAGTCCGCCATGCCCGAAGGCTACGTCCAGGAGCGCACCATCTTCTTCAAGGACGACGGC' + \
          'AACTACAAGACCCGCGCCGAGGTGAAGTTCGAGGGCGACACCCTGGTGAACCGCATCGAGCTG' + \
          'AAGGGCATCGACTTCAAGGAGGACGGCAACATCCTGGGGCACAAGCTGGAGTACAACTACAAC' + \
          'AGCCACAACGTCTATATCATGGCCGACAAGCAGAAGAACGGCATCAAGGTGAACTTCAAGATC' + \
          'CGCCACAACATCGAGGACGGCAGCGTGCAGCTCGCCGACCACTACCAGCAGAACACCCCCATC' + \
          'GGCGACGGCCCCGTGCTGCTGCCCGACAACCACTACCTGAGCTACCAGTCCGCCCTGAGCAAA' + \
          'GACCCCAACGAGAAGCGCGATCACATGGTCCTGCTGGAGTTCGTGACCGCCGCCGGGATCACT' + \
          'CTCGGCATGGACGAGCTGTACAAGTAA'
    dna = DNA(seq)
    prot = 'MVSKGEELFTGVVPILVELDGDVNGHKFSVSGEGEGDATYGKLTLKFICTTGKLPVPWPTLV' + \
           'TTFGYGLQCFARYPDHMKQHDFFKSAMPEGYVQERTIFFKDDGNYKTRAEVKFEGDTLVNRI' + \
           'ELKGIDFKEDGNILGHKLEYNYNSHNVYIMADKQKNGIKVNFKIRHNIEDGSVQLADHYQQN' + \
           'TPIGDGPVLLPDNHYLSYQSALSKDPNEKRDHMVLLEFVTAAGITLGMDELYK'
    rna = reaction.utils.convert_sequence(dna, 'rna')
    r_trans = reaction.utils.convert_sequence(rna, 'dna')
    trans = reaction.utils.convert_sequence(rna, 'peptide')
    assert_equal(str(trans), prot)
    assert_equal(str(r_trans), seq)
    assert_raises(ValueError, reaction.utils.convert_sequence, seq, 'rna')

    # Gapped sequence shouldfail
    assert_raises(ValueError, reaction.utils.convert_sequence,
                  DNA('atg-'), 'rna')

    # Sequence without stop codon should still work
    nostop_dna = DNA('atgaaaaaaaaaaaa')
    nostop_rna = reaction.utils.convert_sequence(nostop_dna, 'rna')
    nostop_peptide = reaction.utils.convert_sequence(nostop_rna, 'peptide')
    assert_equal(str(nostop_rna), 'AUGAAAAAAAAAAAA')
    assert_equal(str(nostop_peptide), 'MKKKK')

    assert_raises(ValueError, reaction.utils.convert_sequence, 'duck', 'rna')



'''
Tests StructureWindows analysis class.

'''

from nose.tools import assert_equal
from coral import analysis, DNA


def test_structure_windows():
    '''Tests StructureWindows class in structure_windows.'''

    seq = 'atggtgagcaagggcgaggagctgttcaccggggtggtgcccatcctggtcgagctggacggc' + \
          'gacgtaaacggccacaagttcagcgtgtccggcgagggcgagggcgatgccacctacggcaag' + \
          'ctgaccctgaagttcatctgcaccaccggcaagctgcccgtgccctggcccaccctcgtgacc' + \
          'accttcggctacggcctgcagtgcttcgcccgctaccccgaccacatgaagcagcacgacttc' + \
          'ttcaagtccgccatgcccgaaggctacgtccaggagcgcaccatcttcttcaaggacgacggc' + \
          'aactacaagacccgcgccgaggtgaagttcgagggcgacaccctggtgaaccgcatcgagctg' + \
          'aagggcatcgacttcaaggaggacggcaacatcctggggcacaagctggagtacaactacaac' + \
          'agccacaacgtctatatcatggccgacaagcagaagaacggcatcaaggtgaacttcaagatc' + \
          'cgccacaacatcgaggacggcagcgtgcagctcgccgaccactaccagcagaacacccccatc' + \
          'ggcgacggccccgtgctgctgcccgacaaccactacctgagctaccagtccgccctgagcaaa' + \
          'gaccccaacgagaagcgcgatcacatggtcctgctggagttcgtgaccgccgccgggatcact' + \
          'ctcggcatggacgagctgtacaagtaa'
    dna_seq = DNA(seq)
    walker = analysis.StructureWindows(dna_seq)
    walker.windows(window_size=60, context_len=90, step=10)
    assert_equal(walker.scores,
                 (0.578570075,
                  0.5928413833333335,
                  0.5535072916666667,
                  0.5425574666666667,
                  0.6028716333333335,
                  0.5907444666666667,
                  0.5532209166666666,
                  0.5882098916666667,
                  0.6471799,
                  0.6957834999999999,
                  0.6209094583333334,
                  0.5929873583333332,
                  0.6117790833333332,
                  0.6116499166666667,
                  0.5987705999999998,
                  0.6439044999999999,
                  0.6817365833333334,
                  0.6488576499999998,
                  0.6900404249999998,
                  0.6657639999999999,
                  0.7083993333333333,
                  0.6360369916666666,
                  0.6452116666666665,
                  0.6395126666666666,
                  0.6288818333333333,
                  0.6351839999999999,
                  0.6463396666666666,
                  0.6717609166666665,
                  0.67853025,
                  0.7012450833333332,
                  0.6620117499999998,
                  0.7250783333333332,
                  0.6995034166666668,
                  0.7386933333333333,
                  0.7494905833333333,
                  0.7247731666666668,
                  0.7510857500000001,
                  0.7458025000000003,
                  0.7434455,
                  0.6702263583333334,
                  0.6390452499999999,
                  0.6503500249999998,
                  0.646285175,
                  0.606586825,
                  0.5707148,
                  0.644573625,
                  0.6644399750000001,
                  0.6716777749999999,
                  0.6807071583333334))



'''Test gibson design module.'''
from nose.tools import assert_equal, assert_raises
from coral import design, DNA, Primer


def test_gibson_primers():
    '''Test gibson_primers function.'''
    # Fuse tdh3 promoter sequence to yfp (trimmed for readability)
    tdh3_3prime = DNA('aaccagttccctgaaattattcccctacttgactaataagtat' +
                      'ataaagacggtaggtattgattgtaattctgtaaatctatttc' +
                      'ttaaacttc')
    yfp_nterm = DNA('atggtgagcaagggcgaggagctgttcaccggggtggtgcccatc' +
                    'ctggtcgagctggacggcgacgtaaacggccacaagttcagcgtg' +
                    'tccggcgagggcgagggcgatgccacctacggcaagctgaccctg' +
                    'aag')
    # Expected annealing sequences and their Tms
    fwd_anneal = DNA('atggtgagcaagggcg')
    fwd_tm = 64.64172107821065
    rev_anneal = DNA('gaagtttaagaaatagatttacagaattacaatcaatac')
    rev_tm = 64.24536287254085
    # Expected overlaps
    all_right = DNA('TCGCCCTTGCTCACCAT')
    all_left = DNA('GGTATTGATTGTAATTCTGTAAATCTATTTCTTAAACTTC')
    mixed_fwd = DNA('TTCTTAAACTTC')
    mixed_rev = DNA('CCTTGCTCACCAT')
    # Design primers - with homology all on left side, right side, or mixed
    # All on the 'right' - i.e. fwd primer
    right = design.gibson_primers(tdh3_3prime, yfp_nterm, 'right')
    right_rev = Primer(rev_anneal, tm=rev_tm, overhang=all_right)
    right_fwd = Primer(fwd_anneal, tm=fwd_tm)
    assert_equal(right, (right_rev, right_fwd))
    # All on the 'left' - i.e. rev primer
    left = design.gibson_primers(tdh3_3prime, yfp_nterm, 'left')
    left_rev = Primer(rev_anneal, tm=rev_tm)
    left_fwd = Primer(fwd_anneal, tm=fwd_tm, overhang=all_left)
    assert_equal(left, (left_rev, left_fwd))
    # On both primers
    mixed = design.gibson_primers(tdh3_3prime, yfp_nterm, 'mixed')
    mixed_primer1 = Primer(rev_anneal, tm=rev_tm, overhang=mixed_rev)
    mixed_primer2 = Primer(fwd_anneal, tm=fwd_tm, overhang=mixed_fwd)
    assert_equal(mixed, (mixed_primer1, mixed_primer2))

    assert_raises(ValueError, design.gibson_primers, tdh3_3prime,
                  yfp_nterm, 'duck')



import coral as cr
import csv
import os
from nose.tools import assert_equal, assert_true
import numpy as np


class TestNUPACK(object):
    def __init__(self):
        self.dnas = [cr.DNA('GATACTAGCG'),
                     cr.DNA('TACGATT'),
                     cr.DNA('GATACG')]
        self.rnas = [s.transcribe() for s in self.dnas]
        self.nupack = cr.analysis.NUPACK()

    def test_pfunc(self):
        '''Test the simplest (partition function) command pfunc with a single
        sequence input.'''
        # test DNA
        output_dna = self.nupack.pfunc(self.dnas[0])
        assert_equal(output_dna, (-4.92069506e-02, 1.08311357973974e+00))
        # test RNA with 1995 params
        output_rna = self.nupack.pfunc(self.rnas[0])
        assert_equal(output_rna, (-9.28516187e-02, 1.16259513934557e+00))
        # test RNA with 1999 params
        output_rna99 = self.nupack.pfunc(self.rnas[0], material='rna1999')
        assert_equal(output_rna99, (-7.97413222e-03, 1.01302234408117e+00))

    def test_pfunc_multi(self):
        '''Test the simplest (partition function) command pfunc with the
        -multi option, which returns an ordered complex partition function.'''
        # test DNA
        output_dna = self.nupack.pfunc_multi(self.dnas)
        assert_equal(output_dna, (-9.59943928e+00, 5.81176347268940e+06))
        # test RNA with 1995 params
        output_rna = self.nupack.pfunc_multi(self.rnas)
        assert_equal(output_rna, (-5.45632785e+00, 6.99579788609535e+03))
        # test RNA with 1999 params
        output_rna99 = self.nupack.pfunc_multi(self.rnas, material='rna1999')
        assert_equal(output_rna99, (-5.27740504e+00, 5.23308895793574e+03))

    def test_pairs(self):
        '''Test the pairs command.'''
        # test DNA
        dna_mat = self._process_ppairs('pairs_dna.tsv', len(self.dnas[0]))
        output_dna = self.nupack.pairs(self.dnas[0])
        assert_true(np.array_equal(dna_mat, output_dna))

        # test RNA
        rna_mat = self._process_ppairs('pairs_rna.tsv', len(self.rnas[0]))
        output_rna = self.nupack.pairs(self.rnas[0])
        assert_true(np.array_equal(rna_mat, output_rna))

        # test RNA 1999
        rna99_mat = self._process_ppairs('pairs_rna99.tsv', len(self.rnas[0]))
        output_rna99 = self.nupack.pairs(self.rnas[0], material='rna1999')
        assert_true(np.array_equal(rna99_mat, output_rna99))

    def test_pairs_multi(self):
        '''Test the pairs command with the -multi flag.'''
        # Test DNA
        dim = sum([len(s) for s in self.dnas])
        dna_ppairs = self._process_ppairs('pairs_multi_dna.ppairs', dim)
        dna_epairs = self._process_ppairs('pairs_multi_dna.epairs', dim)
        dna_output = self.nupack.pairs_multi(self.dnas)
        for expected, output in zip([dna_ppairs, dna_epairs], dna_output):
            assert_true(np.array_equal(expected, output))

        # Test RNA
        dim = sum([len(s) for s in self.rnas])
        rna_ppairs = self._process_ppairs('pairs_multi_rna.ppairs', dim)
        rna_epairs = self._process_ppairs('pairs_multi_rna.epairs', dim)
        rna_output = self.nupack.pairs_multi(self.rnas)
        for expected, output in zip([rna_ppairs, rna_epairs], rna_output):
            assert_true(np.array_equal(expected, output))

        # Test RNA 1999
        dim = sum([len(s) for s in self.rnas])
        rna99_ppairs = self._process_ppairs('pairs_multi_rna99.ppairs', dim)
        rna99_epairs = self._process_ppairs('pairs_multi_rna99.epairs', dim)
        expected_mats = [rna99_ppairs, rna99_epairs]
        rna99_output = self.nupack.pairs_multi(self.rnas, material='rna1999')
        for expected, output in zip(expected_mats, rna99_output):
            assert_true(np.array_equal(expected, output))

    def test_mfe(self):
        # Test DNA
        dna_output = self.nupack.mfe(sum(self.dnas))
        assert_equal(dna_output['mfe'], -1.210)
        assert_equal(dna_output['dotparens'], '........((((.......))))')
        assert_equal(dna_output['pairlist'],
                     [[8, 22], [9, 21], [10, 20], [11, 19]])

        # Test RNA
        rna_output = self.nupack.mfe(sum(self.rnas))
        assert_equal(rna_output['mfe'], -1.100)
        assert_equal(rna_output['dotparens'], '........((((.......))))')
        assert_equal(rna_output['pairlist'],
                     [[8, 22], [9, 21], [10, 20], [11, 19]])

        # Test RNA 1999
        rna99_output = self.nupack.mfe(sum(self.rnas), material='rna1999')
        assert_equal(rna99_output['mfe'], -0.300)
        assert_equal(rna99_output['dotparens'], '........((((.......))))')
        assert_equal(rna99_output['pairlist'],
                     [[8, 22], [9, 21], [10, 20], [11, 19]])

    def test_mfe_degenerate(self):
        # Test '-degenerate' flag with DNA
        degenerate_input = self.dnas[0] + self.dnas[0] + self.dnas[0]
        degenerate_output = self.nupack.mfe(degenerate_input, degenerate=True)
        # Should generate 2 degenerate equal-MFE structures
        assert_equal(degenerate_output[0]['mfe'], -1.330)
        assert_equal(degenerate_output[0]['dotparens'],
                     '..............((((......))))..')
        assert_equal(degenerate_output[0]['pairlist'],
                     [[14, 27], [15, 26], [16, 25], [17, 24]])
        assert_equal(degenerate_output[1]['mfe'], -1.330)
        assert_equal(degenerate_output[1]['dotparens'],
                     '....((((......))))............')
        assert_equal(degenerate_output[1]['pairlist'],
                     [[4, 17], [5, 16], [6, 15], [7, 14]])

    def test_mfe_multi(self):
        # Test DNA
        dna_output = self.nupack.mfe_multi(self.dnas)
        assert_equal(dna_output['mfe'], -8.773)
        assert_equal(dna_output['dotparens'], '.((.....((+..))...+.))...')
        assert_equal(dna_output['pairlist'],
                     [[1, 19], [2, 18], [8, 13], [9, 12]])

        # Test RNA
        rna_output = self.nupack.mfe_multi(self.rnas)
        assert_equal(rna_output['mfe'], -3.863)
        assert_equal(rna_output['dotparens'], '(.......((+..))...+....).')
        assert_equal(rna_output['pairlist'],
                     [[0, 21], [8, 13], [9, 12]])

        # Test RNA 1999
        rna99_output = self.nupack.mfe_multi(self.rnas, material='rna1999')
        assert_equal(rna99_output['mfe'], -4.263)
        assert_equal(rna99_output['dotparens'], '(.......((+..))...+....).')
        assert_equal(rna99_output['pairlist'],
                     [[0, 21], [8, 13], [9, 12]])

    def test_subopt(self):
        # Test DNA
        dna_output = self.nupack.subopt(self.dnas[0], 2.5)
        # For DNA, 3 are found
        assert_equal(dna_output[0]['mfe'], 0.000)
        assert_equal(dna_output[0]['dotparens'], '..........')
        assert_equal(dna_output[0]['pairlist'], [])
        assert_equal(dna_output[1]['mfe'], 1.940)
        assert_equal(dna_output[1]['dotparens'], '....(....)')
        assert_equal(dna_output[1]['pairlist'], [[4, 9]])
        assert_equal(dna_output[2]['mfe'], 2.500)
        assert_equal(dna_output[2]['dotparens'], '.(...)....')
        assert_equal(dna_output[2]['pairlist'], [[1, 5]])

        # Test RNA
        rna_output = self.nupack.subopt(self.rnas[0], 2.5)
        assert_equal(rna_output[0]['mfe'], 0.000)
        assert_equal(rna_output[0]['dotparens'], '..........')
        assert_equal(rna_output[0]['pairlist'], [])
        assert_equal(rna_output[1]['mfe'], 1.300)
        assert_equal(rna_output[1]['dotparens'], '(.......).')
        assert_equal(rna_output[1]['pairlist'], [[0, 8]])

    def test_subopt_multi(self):
        # Test DNA
        dna_output = self.nupack.subopt_multi(self.dnas, 0.5)
        assert_equal(dna_output[0]['mfe'], -8.773)
        assert_equal(dna_output[0]['dotparens'], '.((.....((+..))...+.))...')
        assert_equal(dna_output[0]['pairlist'],
                     [[1, 19], [2, 18], [8, 13], [9, 12]])
        assert_equal(dna_output[1]['mfe'], -8.323)
        assert_equal(dna_output[1]['dotparens'], '.((...(.((+..)).).+.))...')
        assert_equal(dna_output[1]['pairlist'],
                     [[1, 19], [2, 18], [6, 15], [8, 13], [9, 12]])

        # Test RNA
        rna_output = self.nupack.subopt_multi(self.rnas, 0.5)
        assert_equal(rna_output[0]['mfe'], -3.863)
        assert_equal(rna_output[0]['dotparens'], '(.......((+..))...+....).')
        assert_equal(rna_output[0]['pairlist'], [[0, 21], [8, 13], [9, 12]])
        assert_equal(rna_output[1]['mfe'], -3.663)
        assert_equal(rna_output[1]['dotparens'], '.((.....((+..))...+.))...')
        assert_equal(rna_output[1]['pairlist'],
                     [[1, 19], [2, 18], [8, 13], [9, 12]])

        # Test RNA 1999
        rna99_output = self.nupack.subopt_multi(self.rnas, 0.5,
                                                material='rna1999')
        assert_equal(rna99_output[0]['mfe'], -4.263)
        assert_equal(rna99_output[0]['dotparens'],
                     '(.......((+..))...+....).')
        assert_equal(rna99_output[0]['pairlist'], [[0, 21], [8, 13], [9, 12]])

    def test_count(self):
        # Test DNA
        dna_output = self.nupack.count(self.dnas[0])
        assert_equal(dna_output, 9)

        # Test RNA
        rna_output = self.nupack.count(self.rnas[0])
        assert_equal(rna_output, 9)

        # Test DNA
        rna99_output = self.nupack.count(self.rnas[0])
        assert_equal(rna99_output, 9)

    def test_count_multi(self):
        # Test DNA
        dna_output = self.nupack.count_multi(self.dnas)
        assert_equal(dna_output, 7681)

        # Test RNA
        rna_output = self.nupack.count_multi(self.rnas)
        assert_equal(rna_output, 7681)

        # Test RNA 1999
        rna99_output = self.nupack.count_multi(self.rnas)
        assert_equal(rna99_output, 7681)

    def test_energy(self):
        # Test DNA
        dna_output = self.nupack.energy(self.dnas[0], '..(....)..')
        assert_equal(dna_output, 200003.05000000000000)

        # Test RNA
        rna_output = self.nupack.energy(self.rnas[0], '..(....)..')
        assert_equal(rna_output, 200003.70000000000000)

        # Test RNA 1999
        rna99_output = self.nupack.energy(self.rnas[0], '..(....)..',
                                          material='rna1999')
        assert_equal(rna99_output, 200005.59999999999999)

    def test_energy_multi(self):
        # Test DNA
        dna_output = self.nupack.energy_multi(self.dnas,
                                              '(......(((+..))..)+....).')
        assert_equal(dna_output, 199998.83729716184106)

        # Test RNA
        rna_output = self.nupack.energy_multi(self.rnas,
                                              '(......(((+..))..)+....).')
        assert_equal(rna_output, 200002.13729716184108)

        # Test RNA 1999
        rna99_output = self.nupack.energy_multi(self.rnas,
                                                '(......(((+..))..)+....).',
                                                material='rna1999')
        assert_equal(rna99_output, 200002.43729716184109)

    def test_prob(self):
        # Test DNA
        dna_output = self.nupack.prob(self.dnas[0], '..........')
        assert_equal(dna_output, .9233)

        # Test RNA
        rna_output = self.nupack.prob(self.rnas[0], '..........')
        assert_equal(rna_output, .8601)

        # Test RNA 1999
        rna99_output = self.nupack.prob(self.rnas[0], '..........',
                                        material='rna1999')
        assert_equal(rna99_output, .9871)

    def test_prob_multi(self):
        # Test DNA
        dna_output = self.nupack.prob_multi(self.dnas,
                                            '(.......((+..))...+....).')
        assert_equal(dna_output, .04460)

        # Test RNA
        rna_output = self.nupack.prob_multi(self.rnas,
                                            '(.......((+..))...+....).')
        assert_equal(rna_output, .07534)

        # Test RNA 1999
        rna99_output = self.nupack.prob_multi(self.rnas,
                                              '(.......((+..))...+....).',
                                              material='rna1999')
        assert_equal(rna99_output, .1927)

    def test_defect(self):
        # Test DNA
        dna_output = self.nupack.defect(self.dnas[0], '..(....)..')
        assert_equal(dna_output, [2.152, .2152])

        # Test RNA
        rna_output = self.nupack.defect(self.rnas[0], '..(....)..')
        assert_equal(rna_output, [2.274, .2274])

        # Test RNA 1999
        rna99_output = self.nupack.defect(self.rnas[0], '..(....)..',
                                          material='rna1999')
        assert_equal(rna99_output, [2.025, .2025])

    def test_defect_multi(self):
        # Test DNA
        dna_output = self.nupack.defect_multi(self.dnas,
                                              '(.......((+..))...+....).')
        assert_equal(dna_output, [5.790, .2517])

        # Test RNA
        rna_output = self.nupack.defect_multi(self.rnas,
                                              '(.......((+..))...+....).')
        assert_equal(rna_output, [6.522, .2836])

        # Test RNA 1999
        rna99_output = self.nupack.defect_multi(self.rnas,
                                                '(.......((+..))...+....).',
                                                material='rna1999')
        assert_equal(rna99_output, [4.733, .2058])

    def test_complexes(self):
        dnas = self.dnas[:2]
        rnas = self.rnas[:2]
        # Test DNA
        dna_data = [[1, 0, -4.92069506e-02],
                    [0, 1, -2.01874269e-02],
                    [2, 0, -5.76962013e+00],
                    [1, 1, -5.60626976e+00],
                    [0, 2, -6.02915496e+00],
                    [3, 0, -1.02261804e+01],
                    [2, 1, -1.08277874e+01],
                    [1, 2, -9.83335621e+00],
                    [0, 3, -8.94565103e+00],
                    [4, 0, -1.54019941e+01],
                    [3, 1, -1.57681947e+01],
                    [2, 2, -1.56221971e+01],
                    [1, 3, -1.48468888e+01],
                    [0, 4, -1.41217372e+01]]
        dna_expected = [{'complex': [int(row[0]), int(row[1])],
                         'energy': float(row[2])} for row in dna_data]

        dna_output = self.nupack.complexes(dnas, 4)
        for expect, output in zip(dna_expected, dna_output):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])

        # Test RNA
        rna_data = [[1, 0, -9.28516187e-02],
                    [0, 1, -4.32138317e-03],
                    [2, 0, -5.33244849e+00],
                    [1, 1, -2.96200638e+00],
                    [0, 2, -3.28345221e+00],
                    [3, 0, -8.66644413e+00],
                    [2, 1, -8.12678954e+00],
                    [1, 2, -5.50808307e+00],
                    [0, 3, -4.51692546e+00],
                    [4, 0, -1.33713694e+01],
                    [3, 1, -1.17159283e+01],
                    [2, 2, -1.08018186e+01],
                    [1, 3, -7.97713466e+00],
                    [0, 4, -6.98583185e+00]]

        rna_expected = [{'complex': [int(row[0]), int(row[1])],
                         'energy': float(row[2])} for row in rna_data]

        rna_output = self.nupack.complexes(rnas, 4)
        for expect, output in zip(rna_expected, rna_output):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])

        # Test RNA 1999
        rna99_data = [[1, 0, -7.97413222e-03],
                      [0, 1, -1.09376608e-04],
                      [2, 0, -5.66294745e+00],
                      [1, 1, -3.00357694e+00],
                      [0, 2, -3.61831003e+00],
                      [3, 0, -8.58329727e+00],
                      [2, 1, -8.40127714e+00],
                      [1, 2, -5.24448000e+00],
                      [0, 3, -3.89211467e+00],
                      [4, 0, -1.34538247e+01],
                      [3, 1, -1.15873619e+01],
                      [2, 2, -1.07578288e+01],
                      [1, 3, -7.43779859e+00],
                      [0, 4, -6.56448976e+00]]

        rna99_expected = [{'complex': [int(row[0]), int(row[1])],
                           'energy': float(row[2])} for row in rna99_data]

        rna99_output = self.nupack.complexes(rnas, 4, material='rna1999')
        for expect, output in zip(rna99_expected, rna99_output):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])

        # Test DNA with pairs option
        dim = sum([len(x) for x in dnas])
        dnapairs_data = self._read_cxepairs('complexes_pairs_dna.cx-epairs')
        for i, pairlist in enumerate(dnapairs_data):
            dna_expected[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        dnapairs_output = self.nupack.complexes(dnas, 4, pairs=True)
        # Since there's a numpy matrix in there, have to use numpy comparison
        # on each element
        for expected, output in zip(dna_expected, dnapairs_output):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test RNA with pairs option
        dim = sum([len(x) for x in rnas])
        rnapairs_data = self._read_cxepairs('complexes_pairs_rna.cx-epairs')
        for i, pairlist in enumerate(rnapairs_data):
            rna_expected[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        rnapairs_output = self.nupack.complexes(rnas, 4, pairs=True)
        # Since there's a numpy matrix in there, have to use numpy comparison
        # on each element
        for expected, output in zip(rna_expected, rnapairs_output):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test RNA 1999 with pairs option
        dim = sum([len(x) for x in rnas])
        rna99pairs_dat = self._read_cxepairs('complexes_pairs_rna99.cx-epairs')
        for i, pairlist in enumerate(rna99pairs_dat):
            rna99_expected[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        rna99pairs_output = self.nupack.complexes(rnas, 4, pairs=True,
                                                  material='rna1999')
        # Since there's a numpy matrix in there, have to use numpy comparison
        # on each element
        for expected, output in zip(rna99_expected, rna99pairs_output):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test DNA with the ordered option
        dna_ocx = [[1, 0, -4.92069506e-02],
                   [0, 1, -2.01874269e-02],
                   [2, 0, -5.76962013e+00],
                   [1, 1, -5.60626976e+00],
                   [0, 2, -6.02915496e+00],
                   [3, 0, -1.02261804e+01],
                   [2, 1, -1.08277874e+01],
                   [1, 2, -9.83335621e+00],
                   [0, 3, -8.94565103e+00],
                   [4, 0, -1.54019941e+01],
                   [3, 1, -1.57681947e+01],
                   [2, 2, -1.51522348e+01],
                   [2, 2, -1.52349820e+01],
                   [1, 3, -1.48468888e+01],
                   [0, 4, -1.41217372e+01]]
        dna_ocx_expect = [{'complex': [l[0], l[1]], 'energy': l[2]} for l in
                          dna_ocx]
        dna_ocx_keys = [[1],
                        [2],
                        [1, 1],
                        [1, 2],
                        [2, 2],
                        [1, 1, 1],
                        [1, 1, 2],
                        [1, 2, 2],
                        [2, 2, 2],
                        [1, 1, 1, 1],
                        [1, 1, 1, 2],
                        [1, 1, 2, 2],
                        [1, 2, 1, 2],
                        [1, 2, 2, 2],
                        [2, 2, 2, 2]]
        for i, key in enumerate(dna_ocx_keys):
            dna_ocx_expect[i]['order'] = key
        dna_ocx = self.nupack.complexes(dnas, 4, ordered=True)
        for expect, output in zip(dna_ocx_expect, dna_ocx):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])
            assert_equal(expect['order'], output['order'])

        # Test RNA with the ordered option
        rna_ocx = [[1, 0, -9.28516187e-02],
                   [0, 1, -4.32138317e-03],
                   [2, 0, -5.33244849e+00],
                   [1, 1, -2.96200638e+00],
                   [0, 2, -3.28345221e+00],
                   [3, 0, -8.66644413e+00],
                   [2, 1, -8.12678954e+00],
                   [1, 2, -5.50808307e+00],
                   [0, 3, -4.51692546e+00],
                   [4, 0, -1.33713694e+01],
                   [3, 1, -1.17159283e+01],
                   [2, 2, -1.05784008e+01],
                   [2, 2, -1.00680851e+01],
                   [1, 3, -7.97713466e+00],
                   [0, 4, -6.98583185e+00]]
        rna_ocx_expect = [{'complex': [l[0], l[1]], 'energy': l[2]} for l in
                          rna_ocx]
        rna_ocx_keys = [[1],
                        [2],
                        [1, 1],
                        [1, 2],
                        [2, 2],
                        [1, 1, 1],
                        [1, 1, 2],
                        [1, 2, 2],
                        [2, 2, 2],
                        [1, 1, 1, 1],
                        [1, 1, 1, 2],
                        [1, 1, 2, 2],
                        [1, 2, 1, 2],
                        [1, 2, 2, 2],
                        [2, 2, 2, 2]]
        for i, key in enumerate(rna_ocx_keys):
            rna_ocx_expect[i]['order'] = key
        rna_ocx = self.nupack.complexes(rnas, 4, ordered=True)
        for expect, output in zip(rna_ocx_expect, rna_ocx):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])
            assert_equal(expect['order'], output['order'])

        # Test RNA 99 with the ordered option
        rna99_ocx = [[1, 0, -7.97413222e-03],
                     [0, 1, -1.09376608e-04],
                     [2, 0, -5.66294745e+00],
                     [1, 1, -3.00357694e+00],
                     [0, 2, -3.61831003e+00],
                     [3, 0, -8.58329727e+00],
                     [2, 1, -8.40127714e+00],
                     [1, 2, -5.24448000e+00],
                     [0, 3, -3.89211467e+00],
                     [4, 0, -1.34538247e+01],
                     [3, 1, -1.15873619e+01],
                     [2, 2, -1.03498891e+01],
                     [2, 2, -1.03107451e+01],
                     [1, 3, -7.43779859e+00],
                     [0, 4, -6.56448976e+00]]
        rna99_ocx_expect = [{'complex': [l[0], l[1]], 'energy': l[2]} for l
                            in rna99_ocx]
        rna99_ocx_keys = [[1],
                          [2],
                          [1, 1],
                          [1, 2],
                          [2, 2],
                          [1, 1, 1],
                          [1, 1, 2],
                          [1, 2, 2],
                          [2, 2, 2],
                          [1, 1, 1, 1],
                          [1, 1, 1, 2],
                          [1, 1, 2, 2],
                          [1, 2, 1, 2],
                          [1, 2, 2, 2],
                          [2, 2, 2, 2]]
        for i, key in enumerate(rna99_ocx_keys):
            rna99_ocx_expect[i]['order'] = key
        rna99_ocx = self.nupack.complexes(rnas, 4, ordered=True,
                                          material='rna1999')
        for expect, output in zip(rna99_ocx_expect, rna99_ocx):
            assert_equal(expect['complex'], output['complex'])
            assert_equal(expect['energy'], output['energy'])
            assert_equal(expect['order'], output['order'])

        # Test DNA with the ordered and pairs options
        dnapairs_ocx_d = self._read_cxepairs('complexes_pairs_dna.ocx-epairs')
        for i, pairlist in enumerate(dnapairs_ocx_d):
            dna_ocx_expect[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        dna_ocx = self.nupack.complexes(dnas, 4, ordered=True, pairs=True)
        for expected, output in zip(dna_ocx_expect, dna_ocx):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['order'], output['order'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test RNA with the ordered and pairs options
        rnapairs_ocx_d = self._read_cxepairs('complexes_pairs_rna.ocx-epairs')
        for i, pairlist in enumerate(rnapairs_ocx_d):
            rna_ocx_expect[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        rna_ocx = self.nupack.complexes(rnas, 4, ordered=True, pairs=True)
        for expected, output in zip(rna_ocx_expect, rna_ocx):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['order'], output['order'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test RNA 1999 with the ordered and pairs options
        r99pairs_ocx_epairs_file = 'complexes_pairs_rna99.ocx-epairs'
        rna99pairs_ocx_d = self._read_cxepairs(r99pairs_ocx_epairs_file)
        for i, pairlist in enumerate(rna99pairs_ocx_d):
            rna99_ocx_expect[i]['epairs'] = self._pairs_to_np(pairlist, dim)
        rna99_ocx = self.nupack.complexes(rnas, 4, ordered=True, pairs=True,
                                          material='rna1999')
        for expected, output in zip(rna99_ocx_expect, rna99_ocx):
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['order'], output['order'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # Test DNA with the mfe option
        dna_ocx_mfe_expect = self._process_mfe('complexes_mfe_dna.ocx-mfe')
        for expect, mfedat in zip(dna_ocx_expect, dna_ocx_mfe_expect):
            expect['mfe'] = mfedat['mfe']
            expect['dotparens'] = mfedat['dotparens']
            expect['pairlist'] = mfedat['pairlist']
        dna_ocx_mfe = self.nupack.complexes(dnas, 4, mfe=True)
        for expected, output in zip(dna_ocx_expect, dna_ocx_mfe):
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['order'], output['order'])
            assert_equal(expected['mfe'], output['mfe'])
            assert_equal(expected['dotparens'], output['dotparens'])
            assert_equal(expected['pairlist'], output['pairlist'])

        # Test RNA with the mfe option
        rna_ocx_mfe_expect = self._process_mfe('complexes_mfe_rna.ocx-mfe')
        for expect, mfedat in zip(rna_ocx_expect, rna_ocx_mfe_expect):
            expect['mfe'] = mfedat['mfe']
            expect['dotparens'] = mfedat['dotparens']
            expect['pairlist'] = mfedat['pairlist']
        rna_ocx_mfe = self.nupack.complexes(rnas, 4, mfe=True)
        for expected, output in zip(rna_ocx_expect, rna_ocx_mfe):
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['order'], output['order'])
            assert_equal(expected['mfe'], output['mfe'])
            assert_equal(expected['dotparens'], output['dotparens'])
            assert_equal(expected['pairlist'], output['pairlist'])

        # Test RNA 1999 with the mfe option
        rna99_ocx_mfe_expect = self._process_mfe('complexes_mfe_rna99.ocx-mfe')
        for expect, mfedat in zip(rna99_ocx_expect, rna99_ocx_mfe_expect):
            expect['mfe'] = mfedat['mfe']
            expect['dotparens'] = mfedat['dotparens']
            expect['pairlist'] = mfedat['pairlist']
        rna99_ocx_mfe = self.nupack.complexes(rnas, 4, mfe=True,
                                              material='rna1999')
        for expected, output in zip(rna99_ocx_expect, rna99_ocx_mfe):
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['order'], output['order'])
            assert_equal(expected['mfe'], output['mfe'])
            assert_equal(expected['dotparens'], output['dotparens'])
            assert_equal(expected['pairlist'], output['pairlist'])

        # Test DNA with the mfe and pairs options
        dna_ocx_mfe_pairs = self.nupack.complexes(dnas, 4, mfe=True,
                                                  pairs=True)
        for expected, output in zip(dna_ocx_expect, dna_ocx_mfe_pairs):
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['order'], output['order'])
            assert_equal(expected['mfe'], output['mfe'])
            assert_equal(expected['dotparens'], output['dotparens'])
            assert_equal(expected['pairlist'], output['pairlist'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

        # TODO: Restore the 'RNA' material version - for whatever reason, it
        # has a disagreeing dotparens structure in it.

        # Test RNA 1999 with the mfe and pairs options
        rna99_ocx_mfe_pairs = self.nupack.complexes(rnas, 4, mfe=True,
                                                    pairs=True,
                                                    material='rna1999')
        for expected, output in zip(rna99_ocx_expect, rna99_ocx_mfe_pairs):
            assert_equal(expected['energy'], output['energy'])
            assert_equal(expected['complex'], output['complex'])
            assert_equal(expected['order'], output['order'])
            assert_equal(expected['mfe'], output['mfe'])
            assert_equal(expected['dotparens'], output['dotparens'])
            assert_equal(expected['pairlist'], output['pairlist'])
            assert_true(np.array_equal(expected['epairs'], output['epairs']))

    def test_complexes_timeonly(self):
        # Test complex size of 4
        dna_4 = self.nupack.complexes_timeonly(self.dnas[:2], 4)
        assert_equal(dna_4, 0.33)
        rna_4 = self.nupack.complexes_timeonly(self.rnas[:2], 4)
        assert_equal(rna_4, 0.33)
        rna99_4 = self.nupack.complexes_timeonly(self.dnas[:2], 4)
        assert_equal(rna99_4, 0.33)

        # Test complex size of 8
        dna_8 = self.nupack.complexes_timeonly(self.dnas[:2], 8)
        assert_equal(dna_8, 18.66)
        rna_8 = self.nupack.complexes_timeonly(self.rnas[:2], 8)
        assert_equal(rna_8, 18.66)
        rna99_8 = self.nupack.complexes_timeonly(self.dnas[:2], 8)
        assert_equal(rna99_8, 18.66)

    def _test_concentrations(self):
        dnas = self.dnas[:2]

        # Test simple case
        expected = [[9.999907e-07],
                    [9.998446e-07],
                    [7.675362e-11],
                    [3.682595e-12],
                    [1.893207e-12],
                    [2.675905e-16],
                    [1.287201e-16],
                    [2.122290e-18],
                    [4.907111e-19],
                    [8.627347e-21],
                    [6.789014e-22],
                    [1.778772e-22],
                    [2.099592e-24],
                    [4.853486e-25]]

        complexes = self.nupack.complexes(dnas, 4)
        arg_output = self.nupack.concentrations(complexes, 1e-6)
        for expect, out in zip(expected, arg_output):
            assert_equal(expect, out['concentration'])

        # Test ordered case
        ordered_expected = [[9.999907e-07],
                            [9.998446e-07],
                            [7.675362e-11],
                            [3.682595e-12],
                            [1.893207e-12],
                            [2.675905e-16],
                            [1.287201e-16],
                            [2.122290e-18],
                            [4.907111e-19],
                            [8.627347e-21],
                            [6.789014e-22],
                            [1.237905e-22],
                            [5.408664e-23],
                            [2.099592e-24],
                            [4.853486e-25]]

        ocomplexes = self.nupack.complexes(dnas, 4)
        ordered_output = self.nupack.concentrations(ocomplexes, 1e-6,
                                                    ordered=True)
        for expect, out in zip(ordered_expected, ordered_output):
            assert_equal(expect, out['concentration'])

        # Test with pairs
        pairs_expected = self._read_pairs('concentrations_pairs.cx-epairs')
        pairs_output = self.nupack.concentrations(dnas, 1e-6, pairs=True)
        for expect, out in zip(pairs_expected, pairs_output):
            assert_equal(expect, out['concentration'])

        # Test with pairs with complexes as argument
        arg_pairs_output = self.nupack.concentrations(dnas, 1e-6, pairs=True,
                                                      complexes=complexes)
        for expect, out in zip(pairs_expected, arg_pairs_output):
            assert_equal(expect, out['concentration'])

        # Test with pairs and high cutoff
        pairs_ct_expected = [[9.933000e-01],
                             [9.868000e-01],
                             [9.999984e-01],
                             [9.843713e-01],
                             [9.932675e-01],
                             [9.885687e-01],
                             [9.980516e-01],
                             [9.922003e-01],
                             [9.965996e-01],
                             [9.963997e-01],
                             [9.999922e-01],
                             [9.999922e-01],
                             [9.965996e-01],
                             [9.978997e-01],
                             [9.984999e-01]]
        pairs_ct_output = self.nupack.concentrations(dnas, 1e-6, pairs=True,
                                                     cutoff=0.9)
        for expect, out in zip(pairs_ct_expected, pairs_ct_output):
            assert_equal(expect, out['fpairs'])

        # TODO: sort options

    def test_distributions(self):
        dnas = self.dnas[:2]
        # Test with no optional arguments
        dist_cx_expected = [[0, 1],
                            [1, 0],
                            [0, 2],
                            [1, 1],
                            [1, 2],
                            [2, 0],
                            [3, 0],
                            [2, 1],
                            [0, 3],
                            [4, 0],
                            [3, 1],
                            [2, 2],
                            [1, 3]]
        dist_evs_expected = [2.000000e+00,
                             1.000000e+00,
                             1.223070e-14,
                             6.288671e-15,
                             1.170654e-29,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00,
                             0.000000e+00]

        dist_probcols_expected = [[1.210143e-14, 6.288671e-15, 1.000000e+00],
                                  [6.217249e-15, 1.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 1.223070e-14, 0.000000e+00],
                                  [1.000000e+00, 6.288671e-15, 0.000000e+00],
                                  [1.000000e+00, 1.170654e-29, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00],
                                  [1.000000e+00, 0.000000e+00, 0.000000e+00]]

        # complexes = self.nupack.complexes(dnas, 4)
        # output = self.nupack.distributions(complexes, [1, 2], 1e-6)
        # for i, out in enumerate(output):
        #     assert_equal(out, {'complex': dist_cx_expected,
        #                        'ev': dist_evs_expected,
        #                        'probcols': dist_probcols_expected})

    def _process_mfe(self, filename):
        curdir = os.path.dirname(__file__)
        mfepath = os.path.join(curdir, 'data', filename)
        with open(mfepath) as f:
            mfefile = f.read()
        commentline = '% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %'
        sections = mfefile.split(commentline)
        output = []
        for section in sections[1::2]:
            lines = section.split('\n')
            # Remove first three lines - information not used
            lines.pop(0)
            lines.pop(0)
            lines.pop(0)
            # Remove the last (empty) line
            lines.pop()
            mfe_data = {}
            mfe_data['mfe'] = float(lines[0])
            mfe_data['dotparens'] = lines[1]
            pairlist = []
            if len(lines) > 2:
                for line in lines[2:]:
                    data = line.split('\t')
                    pairlist.append([int(data[0]) - 1, int(data[1]) - 1])
            mfe_data['pairlist'] = pairlist
            output.append(mfe_data)

        return output

    def _read_cxepairs(self, filename):
        curdir = os.path.dirname(__file__)
        epairspath = os.path.join(curdir, 'data', filename)
        with open(epairspath) as f:
            epairsfile = f.read()
        commentline = '% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %'
        sections = epairsfile.split(commentline)
        output = []
        for section in sections[1::2]:
            # Skipping every other one eliminates comments and blank lines
            lines = section.split('\n')
            # Remove first three lines - information not used
            lines.pop(0)
            lines.pop(0)
            lines.pop(0)
            # Remove the last (empty) line
            lines.pop()
            # Extract data
            complex_data = []
            for line in lines:
                data = line.split('\t')
                base_i = int(data[0])
                base_j = int(data[1])
                prob = float(data[2])
                complex_data.append((base_i, base_j, prob))
            output.append(complex_data)

        return output

    def _process_ppairs(self, filename, dim):
        curdir = os.path.dirname(__file__)
        tsvpath = os.path.join(curdir, 'data', filename)
        with open(tsvpath) as f:
            pairlist = [row for row in csv.reader(f, delimiter='\t')]

        return self._pairs_to_np(pairlist, dim)

    def _pairs_to_np(self, pairlist, dim):
        '''Given a set of pair probability lines, construct a numpy array.

        :param pairlist: a list of pair probability triples
        :type pairlist: list

        '''
        mat = np.zeros((dim, dim + 1))
        for line in pairlist:
            i = int(line[0]) - 1
            j = int(line[1]) - 1
            prob = float(line[2])
            mat[i, j] = prob
        return mat



'''
Tests for Needleman-Wunsch function 'needle'

'''

from nose.tools import assert_equal
from coral import analysis, DNA


def test_needle():
    ref_seq = DNA("ATGCGATACGATA")

    res_seq0 = DNA("ATGCGATA---TA")  # Gapped
    res_seq1 = DNA("ATGCGATAATGCGATA")  # Insertion
    res_seq2 = DNA("ATGCGATATA")  # Deletion
    res_seq3 = DNA("ATGCGATAAGATA")  # Mismatch
    results = [res_seq0, res_seq1, res_seq2, res_seq3]

    exp_seq0 = (DNA("ATGCGATACGATA"), DNA("ATGCGATA---TA"), 9)
    exp_seq1 = (DNA("ATGCGATA---CGATA"), DNA("ATGCGATAATGCGATA"), 12)
    exp_seq2 = (DNA("ATGCGATACGATA"), DNA("ATGCGATA---TA"), 9)
    exp_seq3 = (DNA("ATGCGATACGATA"), DNA("ATGCGATAAGATA"), 11)

    expected = [exp_seq0, exp_seq1, exp_seq2, exp_seq3]

    for seq, exp in zip(results, expected):
        aligned = analysis.needle(ref_seq, seq, gap_open=-1, gap_extend=0)
        assert_equal(aligned, exp)



'''
Tests for the OligoAssembly design class.

'''

from nose.tools import assert_equal, assert_raises, assert_true
from coral import design, DNA


def test_oligo_assembly():
    '''
    Tests output of OligoAssembly class.

    '''

    seq = 'atgcgtaaaggagaagaacttttcactggagttgtcccaattcttgttgaattagatggtgat' + \
          'gttaatgggcacaaattttctgtcagtggagagggtgaaggtgatgcaacatacggaaaactt' + \
          'acccttaaatttatttgcactactggaaaactacctgttccatggccaacacttgtcactact' + \
          'ttcggttatggtgttcaatgctttgcgagatacccagatcatatgaaacagcatgactttttc' + \
          'aagagtgccatgcccgaaggttatgtacaggaaagaactatatttttcaaagatgacgggaac' + \
          'tacaagacacgtgctgaagtcaagtttgaaggtgatacccttgttaatagaatcgagttaaaa' + \
          'ggtattgattttaaagaagatggaaacattcttggacacaaattggaatacaactataactca' + \
          'cacaatgtatacatcatggcagacaaacaaaagaatggaatcaaagttaacttcaaaattaga' + \
          'cacaacattgaagatggaagcgttcaactagcagaccattatcaacaaaatactccaattggc' + \
          'gatggccctgtccttttaccagacaaccattacctgtccacacaatctgccctttcgaaagat' + \
          'cccaacgaaaagagagaccacatggtccttcttgagtttgtaacagctgctgggattacacat' + \
          'ggcatggatgaactatacaaaaggcctgctgcaaacgacgaaaactacgctttagtagcttaa'
    dna_seq = DNA(seq)

    # Test to make sure oligo_number parameter is working
    reference_oligos = ['atgcgtaaaggagaagaacttttcactggagttgtcccaattcttgttg' +
                        'aattagatggtgatgttaatgggcacaaattttctgtcagtgg',
                        'tttccagtagtgcaaataaatttaagggtaagttttccgtatgttgcat' +
                        'caccttcaccctctccactgacagaaaatttgtgcccattaaca',
                        'cggaaaacttacccttaaatttatttgcactactggaaaactacctgtt' +
                        'ccatggccaacacttgtcactactttcggttatggtgttcaatgc',
                        'ttcgggcatggcactcttgaaaaagtcatgctgtttcatatgatctggg' +
                        'tatctcgcaaagcattgaacaccataaccgaaagtagtgacaagt',
                        'ttcaagagtgccatgcccgaaggttatgtacaggaaagaactatatttt' +
                        'tcaaagatgacgggaactacaagacacgtgctgaagtcaagtttgaa',
                        'agaatgtttccatcttctttaaaatcaataccttttaactcgattctat' +
                        'taacaagggtatcaccttcaaacttgacttcagcacgtgtcttgtagt',
                        'tcgagttaaaaggtattgattttaaagaagatggaaacattcttggaca' +
                        'caaattggaatacaactataactcacacaatgtatacatcatggcaga',
                        'tcttcaatgttgtgtctaattttgaagttaactttgattccattctttt' +
                        'gtttgtctgccatgatgtatacattgtgtgagttatagttgtattc',
                        'ggaatcaaagttaacttcaaaattagacacaacattgaagatggaagcg' +
                        'ttcaactagcagaccattatcaacaaaatactccaattggcgatgg',
                        'cgttgggatctttcgaaagggcagattgtgtggacaggtaatggttgtc' +
                        'tggtaaaaggacagggccatcgccaattggagtattttgttga',
                        'tgccctttcgaaagatcccaacgaaaagagagaccacatggtccttctt' +
                        'gagtttgtaacagctgctgggattacacatggcat',
                        'ttaagctactaaagcgtagttttcgtcgtttgcagcaggccttttgtat' +
                        'agttcatccatgccatgtgtaatcccagcagctg']

    oligo_n_assembly = design.OligoAssembly(dna_seq,
                                            tm=72,
                                            length_range=(100, 160),
                                            require_even=True,
                                            start_5=True,
                                            oligo_number=12)
    oligo_n_assembly.design_assembly()

    oligo_n_output = [str(oligo).lower() for oligo in oligo_n_assembly.oligos]
    assert_equal(oligo_n_output, reference_oligos)

    # Test to make sure oligo_number parameter fails with too restrictive of
    # settings
    def design_impossible(test_seq):
        assembly = design.OligoAssembly(test_seq,
                                        tm=72,
                                        length_range=(120, 120),
                                        require_even=True,
                                        start_5=True,
                                        oligo_number=12)
        assembly.design_assembly()

    assert_raises(Exception, design_impossible, dna_seq)


def test_overlapping_overlaps():
    '''
    Sometimes, an assembly produces a result with 'overlapping overlaps' -
    not ideal. This should eventually be replaced by a catchable exception or
    prevented outright.

    '''

    test_seq = DNA('ATCAATACTTATTACGATATATATAT' * 34)
    oligo_n_assembly = design.OligoAssembly(test_seq,
                                            tm=65,
                                            length_range=(80, 150),
                                            require_even=True,
                                            start_5=True,
                                            overlap_min=20,
                                            oligo_number=10)
    oligo_n_assembly.design_assembly()
    assert_true(type(oligo_n_assembly.warning) == str)



'''Needleman-Wunsch alignment functions.'''
import coral as cr
from . import substitution_matrices as submat
import multiprocessing
import warnings
try:
    from .calign import aligner, score_alignment
except ImportError:
    message = ('NW alignment extension could not be imported, falling back'
               'on native Python version (~100 times slower).')
    warnings.warn(message)
    from .align import aligner, score_alignment


def needle(reference, query, gap_open=-15, gap_extend=0,
           matrix=submat.DNA_SIMPLE):
    '''Do a Needleman-Wunsch alignment.

    :param reference: Reference sequence.
    :type reference: coral.DNA
    :param query: Sequence to align against the reference.
    :type query: coral.DNA
    :param gapopen: Penalty for opening a gap.
    :type gapopen: float
    :param gapextend: Penalty for extending a gap.
    :type gapextend: float
    :param matrix: Matrix to use for alignment - options are DNA_simple (for
                   DNA) and BLOSUM62 (for proteins).
    :type matrix: str
    :returns: (aligned reference, aligned query, score)
    :rtype: tuple of two coral.DNA instances and a float

    '''
    # Align using cython Needleman-Wunsch
    aligned_ref, aligned_res = aligner(str(reference),
                                       str(query),
                                       gap_open=gap_open,
                                       gap_extend=gap_extend,
                                       method='global_cfe',
                                       matrix=matrix)

    # Score the alignment
    score = score_alignment(aligned_ref, aligned_res, gap_open, gap_extend,
                            matrix)

    return cr.DNA(aligned_ref), cr.DNA(aligned_res), score


def run_needle(args):
    '''Run needle command using 4-tuple of the arguments (in the same order)
    as is used for needle. Necessary to make picklable function for
    multiprocessing.'''
    return needle(*args)


def needle_msa(reference, results, gap_open=-15, gap_extend=0,
               matrix=submat.DNA_SIMPLE):
    '''Create a multiple sequence alignment based on aligning every result
    sequence against the reference, then inserting gaps until every aligned
    reference is identical

    '''
    gap = '-'
    # Convert alignments to list of strings
    alignments = []
    for result in results:
        ref_dna, res_dna, score = needle(reference, result, gap_open=gap_open,
                                         gap_extend=gap_extend, matrix=matrix)
        alignments.append([str(ref_dna), str(res_dna), score])

    def insert_gap(sequence, position):
        return sequence[:position] + gap + sequence[position:]

    i = 0
    while True:
        # Iterate over 'columns' in every reference
        refs = [alignment[0][i] for alignment in alignments]

        # If there's a non-unanimous gap, insert gap into alignments
        gaps = [ref == gap for ref in refs]
        if any(gaps) and not all(gaps):
            for alignment in alignments:
                if alignment[0][i] != gap:
                    alignment[0] = insert_gap(alignment[0], i)
                    alignment[1] = insert_gap(alignment[1], i)

        # If all references match, we're all done
        alignment_set = set(alignment[0] for alignment in alignments)
        if len(alignment_set) == 1:
            break

        # If we've reach the end of some, but not all sequences, add end gap
        lens = [len(alignment[0]) for alignment in alignments]
        if i + 1 in lens:
            for alignment in alignments:
                if len(alignment[0]) == i + 1:
                    alignment[0] = alignment[0] + gap
                    alignment[1] = alignment[1] + gap

        i += 1

        if i > 20:
            break

    # Convert into MSA format
    output_alignment = [cr.DNA(alignments[0][0])]
    for alignment in alignments:
        output_alignment.append(cr.DNA(alignment[1]))

    return output_alignment


def needle_multi(references, queries, gap_open=-15, gap_extend=0,
                 matrix=submat.DNA_SIMPLE):
    '''Batch process of sequencing split over several cores. Acts just like
    needle but sequence inputs are lists.

    :param references: References sequence.
    :type references: coral.DNA list
    :param queries: Sequences to align against the reference.
    :type queries: coral.DNA list
    :param gap_open: Penalty for opening a gap.
    :type gap_open: float
    :param gap_extend: Penalty for extending a gap.
    :type gap_extend: float
    :param matrix: Matrix to use for alignment - options are DNA_simple (for
                   DNA) and BLOSUM62 (for proteins).
    :type matrix: str
    :returns: a list of the same output as coral.sequence.needle
    :rtype: list

    '''
    pool = multiprocessing.Pool()
    try:
        args_list = [[ref, que, gap_open, gap_extend, matrix] for ref, que in
                     zip(references, queries)]
        aligned = pool.map(run_needle, args_list)
    except KeyboardInterrupt:
        print 'Caught KeyboardInterrupt, terminating workers'
        pool.terminate()
        pool.join()
        raise KeyboardInterrupt

    return aligned
